# -*- coding: utf-8 -*-
import django

try:
    from importlib.metadata import version
except ImportError:  # pragma: no cover
    from importlib_metadata import version

# Django < 3.2 needs default_app_config in order to correctly register the app
# without specifying the full module path in settings.py
if django.VERSION < (3, 2):  # pragma: no-cover
    default_app_config = "flexible_forms.apps.FlexibleFormsConfig"

try:
    __version__ = version(__name__)
except:
    pass



# -*- coding: utf-8 -*-
from typing import Type, cast

from django.apps import apps
from django.http import HttpRequest, JsonResponse
from django.shortcuts import get_object_or_404

from flexible_forms.cache import cache
from flexible_forms.fields import BaseAutocompleteSelectField
from flexible_forms.models import BaseField, BaseRecord


def autocomplete(
    request: HttpRequest, app_label: str, model_name: str, field_pk: str
) -> JsonResponse:
    """Perform a search and return autocomplete results.

    Args:
        request: The current HTTP request.
        app_label: The app_label of the appropriate BaseField implementation.
        model_name: The app_label of the appropriate BaseField implementation.
        field_pk: The primary key of the Field record.

    Returns:
        JsonResponse: A select2-compatible JSON response of autocomplete suggestions.
    """
    # Retrieve the Field.
    field_model = cast(Type[BaseField], apps.get_model(app_label, model_name))
    field = get_object_or_404(field_model, pk=field_pk)

    # If autocompletion was requested for a specific record, fetch it using
    # the given primary key. If not, use a blank record.
    record_pk = request.GET.get("record_pk")
    record_model = field.flexible_forms.get_model(BaseRecord)
    record = (
        get_object_or_404(record_model, pk=record_pk)
        if record_pk is not None
        else record_model(form=field.form)
    )

    # Fetch the current set of field values from the cache so that the
    # underlying field type can use them to make decisions.
    field_values = cache.get(
        f"flexible_forms:field_values:{record._meta.app_label}:{record._meta.model_name}:{record_pk}",
        default={},
    )

    field_type = cast(
        BaseAutocompleteSelectField,
        field.as_field_type(record=record, field_values=field_values),
    )

    search_results, has_more = field_type.autocomplete(request)

    # Perform the search and return a Select2-compatible response.
    return JsonResponse(
        {
            "results": search_results,
            "pagination": {"more": has_more},
        }
    )



# -*- coding: utf-8 -*-

"""Field definitions for the flexible_forms module."""

import json
import logging
import urllib.parse as urlparse
from datetime import date, datetime
from functools import reduce
from operator import add, ior
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    cast,
)
from urllib.parse import urlencode

import requests
import simpleeval
from django.apps import apps
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import Paginator
from django.core.validators import (
    BaseValidator,
    MaxValueValidator,
    MinValueValidator,
)
from django.db import connections
from django.db.models import Case, F, FileField, ImageField
from django.db.models import Model as DjangoModel
from django.db.models import Sum, When
from django.db.models import fields as model_fields
from django.db.models.query import QuerySet
from django.forms import fields as form_fields
from django.forms import widgets as form_widgets
from django.http import HttpRequest, QueryDict
from django.urls import Resolver404, resolve, reverse
from django.utils.functional import cached_property
from typing_extensions import TypedDict

from flexible_forms.widgets import (
    AutocompleteSelect,
    AutocompleteSelectMultiple,
)

try:
    from django.db.models import JSONField
    from django.forms import JSONField as JSONFormField
except ImportError:  # pragma: no cover
    from django.contrib.postgres.fields import JSONField  # type: ignore
    from django.contrib.postgres.forms import JSONField as JSONFormField  # type: ignore

from flexible_forms.utils import (
    RenderedString,
    check_supports_pg_trgm,
    collect_annotations,
    create_autocomplete_result,
    evaluate_expression,
    get_expression_fields,
    interpolate,
    jp,
)

logger = logging.getLogger(__name__)

##
# _EMPTY_CHOICES
_EMPTY_CHOICES = (("EMPTY", "Select a value."),)

if TYPE_CHECKING:  # pragma: no cover
    from flexible_forms.models import BaseField, BaseRecord


##
# FIELD_TYPES
#
# A dict mapping of field types, where the key is the string representation of
# the field type (usually its class name), and the value is the `FlexibleField`
# class itself.
#
# Built dynamically to include all descendants of `FieldType`.
#
FIELD_TYPES: Dict[str, Type["FieldType"]] = {}


class FieldTypeOptions:
    """A base class for the Meta inner class for FieldType."""

    abstract = False
    force_replacement = False


class FieldTypeMetaclass(type):
    """A metaclass for handling FieldType configuration and registration."""

    label: str

    name: str
    _meta: FieldTypeOptions

    def __new__(
        cls: Type["FieldTypeMetaclass"],
        name: str,
        bases: Tuple[type, ...],
        attrs: Dict[str, Any],
        **kwargs: Any,
    ) -> "FieldTypeMetaclass":
        """Create a new FieldType and register it.

        Args:
            cls: This metaclass.
            name: The name of the new FieldType class.
            bases: The base classes of the new FieldType class.
            attrs: The attributes of the new FieldType class.
            kwargs: Passed to super.

        Returns:
            Type[FieldType]: The new FieldType class, configured and registered.

        Raises:
            ValueError: If a FieldType with the same class name was already
                registered, and the force_replacement meta option is False.
        """
        attrs_meta = attrs.pop("Meta", None)
        attrs["_meta"] = type(
            "Meta", tuple(filter(bool, (attrs_meta, FieldTypeOptions))), {}
        )

        # Set the name from the class name if a name was not provided.
        attrs["name"] = attrs.get("name", name)

        clsobj = super().__new__(cls, name, bases, attrs, **kwargs)

        # Throw an error if a FieldType with the given name was already registered.
        if clsobj.name in FIELD_TYPES and not clsobj._meta.force_replacement:
            raise ValueError(
                f"A FieldType named {name} was already registered ({FIELD_TYPES[name]})."
            )

        # Register the new field type with its class name (if it's not abstract).
        if not clsobj._meta.abstract:
            FIELD_TYPES[name] = cast(Type[FieldType], clsobj)

        return clsobj


class FieldType(metaclass=FieldTypeMetaclass):
    """A prefabricated field to use with flexible forms.

    Provides an interface to emit a model field, form field, or widget.
    """

    # The name is set by the Metaclass when the class is initialized.
    name: str

    class Meta:
        abstract = True

    ##
    # label
    #
    # The human-friendly key of the field type, e.g. "Single-line Text".
    #
    label: str = ""

    ##
    # form_field_class
    #
    # The class to use when creating a new instance of the field for use in a Django
    # form.
    #
    # Default: django.forms.fields.CharField
    #
    form_field_class: Type[form_fields.Field] = form_fields.CharField

    ##
    # form_field_options
    #
    # Keyword arguments to be passed to the form_field_class constructor.
    #
    # Default: {}
    #
    form_field_options: Dict[str, Any] = {}

    ##
    # form_widget_class
    #
    # The class to use to render the form widget to a form template. If unspecified, use
    # the default for the form_field_class.
    #
    # Default: None
    #
    form_widget_class: Optional[Type[form_widgets.Widget]] = None

    ##
    # form_widget_options
    #
    # Keyword arguments to be passed to the form_widget_class constructor.
    #
    # Default: {}
    #
    form_widget_options: Dict[str, Any] = {}

    ##
    # model_field_class
    #
    # The class to use for when using the field to define a Django model.
    #
    # Default: JSONField
    #
    model_field_class: Type[model_fields.Field] = JSONField

    ##
    # model_field_options
    #
    # Keyword arguments to be passed to the model_field_class constructor.
    #
    # Default: {}
    #
    model_field_options: Dict[str, Any] = {}

    def __init__(
        self,
        field: "BaseField",
        record: Optional["BaseRecord"],
        field_values: Optional[Dict[str, Any]] = None,
        modifiers: Sequence[Tuple[str, str]] = (),
        **field_type_options: Any,
    ):
        self.field = field
        self.record = record
        self.field_values = field_values or {}
        self.modifiers = modifiers

        field_type = type(self)
        declared_options = {**collect_annotations(field_type), **field_type.__dict__}
        for attr, value in field_type_options.items():
            if attr not in declared_options:
                raise AttributeError(
                    f"Field type {type(self)} has no attribute {attr}."
                )
            setattr(self, attr, value)

    def as_form_field(self, **form_field_options: Any) -> form_fields.Field:
        """Return an instance of the field for use in a Django form.

        Passes additional kwargs to the form field constructor.

        Args:
            form_field_options: Additional kwargs passed to the Django form
                field constructor.

        Returns:
            form_fields.Field: An instance of the form field.
        """
        # Cascade form field options in precedence order:
        #
        #   * Defaults for the field type.
        #   * Common attributes from the BaseField record.
        #   * Additional options (form_field_options) from the BaseField record.
        #   * Any form_field_options passed to the method directly.
        #
        form_field_options = {
            **self.form_field_options,
            "label": self.field.label,
            "label_suffix": self.field.label_suffix,
            "help_text": self.field.help_text,
            "required": self.field.required,
            "initial": self.field.initial,
            "error_messages": self.field.error_messages,
            "localize": self.field.localize,
            "disabled": self.field.disabled,
            **self.field.form_field_options,
            **form_field_options,
        }

        # If no widget was given explicitly, build a default one.
        form_field_options["widget"] = form_field_options.get(
            "widget", self.as_form_widget()
        )

        # Generate the form field with its appropriate class and widget.
        form_field = self.form_field_class(**form_field_options)

        # Apply any configured modifiers to the field.
        form_field = self.apply_modifiers(form_field)

        return form_field

    def as_form_widget(self, **form_widget_options: Any) -> form_widgets.Widget:
        """Return an instance of the form widget for rendering.

        Passes additional kwargs to the form widget constructor.

        Args:
            form_widget_options: Additional kwargs passed to the Django form
                widget constructor.

        Returns:
            form_widgets.Widget: The configured form widget for the field.
        """
        # If no form_widget_class is configured for the field type, use the
        # default one for the form_field_class.
        widget_cls = self.form_widget_class or cast(
            Type[form_widgets.Widget], self.form_field_class.widget
        )

        return widget_cls(
            **{
                **self.form_widget_options,
                **self.field.form_widget_options,
                **form_widget_options,
            }
        )

    @classmethod
    def as_model_field(cls, **model_field_options: Any) -> model_fields.Field:
        """Return an instance of the field for use in a Django model.

        Receives a dict of kwargs to pass through to the model field constructor.

        Since this method generates a model field (which should only ever be
        used when defining model classes, and should never have dynamic
        attributes), it is a classmethod and does not have access to a field,
        record, or field values.

        Args:
            model_field_options: A dict of kwargs to be passed to the Django
                model field constructor.

        Returns:
            model_fields.Field: An instance of the model field.
        """
        return cls.model_field_class(
            **{
                **cls.model_field_options,
                **model_field_options,
            }
        )

    def apply_modifiers(self, form_field: form_fields.Field) -> form_fields.Field:
        """Apply the given modifiers to the given Django form field.

        Args:
            form_field: The form field to which to apply modifiers.

        Returns:
            form_fields.Field: The given form field, modified using the
                configured modifiers.
        """
        for attribute, expression in self.modifiers:
            expression_context = self.field_values

            if self.record:
                record_variable = (
                    (self.record._meta.verbose_name or "record")
                    .lower()
                    .replace(" ", "_")
                )
                expression_context = {record_variable: self.record, **self.field_values}

            # Evaluate the expression and set the attribute specified by
            # `self.attribute` to the value it returns.
            try:
                expression_value = evaluate_expression(
                    expression, names=expression_context
                )
            except simpleeval.NameNotDefined:
                continue

            # If the caller has implemented a custom apply_ATTRIBUTENAME method
            # to handle application of the attribute, use it.
            custom_applicator = getattr(self, f"apply_{attribute}", None)
            if custom_applicator:
                form_field = custom_applicator(
                    form_field, **{attribute: expression_value}
                )

            # If no custom applicator method is implemented, but the form field
            # has an attribute with the specified name, set its value to the
            # result of the expression.
            elif hasattr(form_field, attribute):
                setattr(form_field, attribute, expression_value)

            # Finally, add the modifier and its value to the applied modifiers
            # dict on the field.
            setattr(
                form_field,
                "_modifiers",
                {
                    **getattr(form_field, "_modifiers", {}),
                    attribute: expression_value,
                },
            )

        return form_field

    def apply_hidden(
        self,
        form_field: form_fields.Field,
        hidden: bool,
        **kwargs: Any,
    ) -> form_fields.Field:
        """Apply the "hidden" attribute.

        If a field modifier specifies that the field should be "hidden", its
        widget is changed to a HiddenInput, and its "required" attribute is
        set to False (to avoid validation errors with fields that can't be
        seen).

        Args:
            form_field: The form field that should be hidden.
            hidden: The new value of "hidden".
            kwargs: Unused.

        Returns:
            form_fields.Field: The given form_field, modified to use the
                given "hidden" value.
        """
        if hidden:
            form_field.widget = form_widgets.HiddenInput()
            form_field.required = False

        return form_field

    def apply_value(
        self,
        form_field: form_fields.Field,
        value: Any,
        **kwargs: Any,
    ) -> form_fields.Field:
        """Apply the "value" field modifier.

        Allows a field modifier to set the value of a field based on an
        expression.

        Args:
            form_field: The form field for which to set the value.
            value: The new value.
            kwargs: Unused.

        Returns:
            form_fields.Field: The given form_field, modified to use the
                given value.
        """
        setattr(form_field, "_value", value)

        return form_field


class SingleLineTextField(FieldType):
    """A field for collecting single-line text values."""

    label = "Single-line Text"

    form_field_class = form_fields.CharField
    model_field_class = model_fields.TextField


class MultiLineTextField(FieldType):
    """A field for collecting multiline text values."""

    label = "Multi-line Text"

    form_field_class = form_fields.CharField
    form_widget_class = form_widgets.Textarea
    model_field_class = model_fields.TextField


class EmailField(FieldType):
    """A field for collecting email addresses."""

    label = "Email Address"

    form_field_class = form_fields.EmailField
    model_field_class = model_fields.EmailField


class URLField(FieldType):
    """A field for collecting URL values."""

    label = "URL"

    form_field_class = form_fields.URLField
    model_field_class = model_fields.URLField
    model_field_options = {"max_length": 2083}


class SensitiveTextField(FieldType):
    """A field for collecting sensitive text values.

    Primarily useful for obfuscating the input when on-screen.
    """

    label = "Sensitive text"

    form_field_class = form_fields.CharField
    form_widget_class = form_widgets.PasswordInput
    model_field_class = model_fields.TextField


class IntegerField(FieldType):
    """A field for collecting integer values."""

    label = "Integer"

    form_field_class = form_fields.IntegerField
    form_field_options = {"min_value": -2147483648, "max_value": 2147483647}
    form_widget_class = form_widgets.NumberInput
    form_widget_options = {"attrs": {"step": "1"}}
    model_field_class = model_fields.IntegerField


class PositiveIntegerField(FieldType):
    """A field for collecting positive integer values."""

    label = "Positive Integer"

    form_field_class = form_fields.IntegerField
    form_field_options = {"min_value": 1, "max_value": 2147483647}
    form_widget_class = form_widgets.NumberInput
    form_widget_options = {"attrs": {"step": "1"}}
    model_field_class = model_fields.PositiveIntegerField


class DecimalField(FieldType):
    """A field for collecting decimal number values."""

    label = "Decimal Number"

    form_field_class = form_fields.DecimalField
    form_field_options = {"max_digits": 15, "decimal_places": 6}
    form_widget_class = form_widgets.NumberInput
    form_widget_options = {"attrs": {"step": "0.01"}}
    model_field_class = model_fields.DecimalField
    model_field_options = {"max_digits": 15, "decimal_places": 6}


class FloatField(FieldType):
    """A field for collecting decimal number values."""

    label = "Floating-point Number"

    form_field_class = form_fields.FloatField
    form_widget_class = form_widgets.NumberInput
    form_widget_options = {"attrs": {"step": "0.1"}}
    model_field_class = model_fields.FloatField


class DateField(FieldType):
    """A field for collecting date data."""

    label = "Date"

    form_field_class = form_fields.DateField
    form_widget_class = form_widgets.DateInput
    form_widget_options = {"attrs": {"type": "date"}}
    model_field_class = model_fields.DateField

    # Field type options that dictate whether the collected date can be in the
    # past or future (both true by default).
    allow_past: bool = True
    allow_future: bool = True

    def as_form_field(self, **form_field_options: Any) -> form_fields.Field:
        """Return a configured form field.

        Adds validators to handle past and future date limitations.

        Args:
            form_field_options: Additional kwargs passed to the Django form
                field constructor.

        Returns:
            form_fields.Field: An instance of the form field.
        """
        return super().as_form_field(
            **{**form_field_options, "validators": self._get_validators()}
        )

    def as_form_widget(self, **form_widget_options: Any) -> form_widgets.Widget:
        """Return an instance of the form widget for rendering.

        Adds widget attributes to handle past and future date limitations.

        Args:
            form_widget_options: Additional kwargs passed to the Django form
                widget constructor.

        Returns:
            form_widgets.Widget: The configured form widget for the field.
        """
        attrs = {
            **self.form_widget_options,
            **self.field.form_widget_options,
            **form_widget_options,
        }.get("attrs", {})

        if not self.allow_past:
            attrs["min"] = date.today().isoformat()

        if not self.allow_future:
            attrs["max"] = date.today().isoformat()

        return super().as_form_widget(**{**form_widget_options, "attrs": attrs})

    def _get_validators(self) -> Tuple[BaseValidator, ...]:
        """Return validators for handling past and future date limitations."""
        validators = cast(Tuple[BaseValidator, ...], ())

        if not self.allow_past:
            validators = (*validators, MinValueValidator(limit_value=date.today))

        if not self.allow_future:
            validators = (*validators, MaxValueValidator(limit_value=date.today))

        return validators


class TimeField(FieldType):
    """A field for collecting time data."""

    label = "Time"

    form_field_class = form_fields.TimeField
    form_widget_class = form_widgets.TimeInput
    form_widget_options = {"attrs": {"type": "time"}}
    model_field_class = model_fields.TimeField


class DateTimeField(FieldType):
    """A field for collecting datetime data."""

    label = "Date & Time"

    form_field_class = form_fields.DateTimeField
    model_field_class = model_fields.DateTimeField

    # Field type options that dictate whether the collected datetime can be in
    # the past or future (both true by default).
    allow_past = True
    allow_future = True

    def as_form_field(self, **form_field_options: Any) -> form_fields.Field:
        """Return a configured form field.

        Adds validators to handle past and future date limitations.

        Args:
            form_field_options: Additional kwargs passed to the Django form
                field constructor.

        Returns:
            form_fields.Field: An instance of the form field.
        """
        return super().as_form_field(
            **{**form_field_options, "validators": self._get_validators()}
        )

    def as_form_widget(self, **form_widget_options: Any) -> form_widgets.Widget:
        """Return an instance of the form widget for rendering.

        Adds widget attributes to handle past and future date limitations.

        Args:
            form_widget_options: Additional kwargs passed to the Django form
                widget constructor.

        Returns:
            form_widgets.Widget: The configured form widget for the field.
        """
        attrs = form_widget_options.get("attrs", {})

        if not self.allow_past:
            attrs["min"] = datetime.now().isoformat()

        if not self.allow_future:
            attrs["max"] = datetime.now().isoformat()

        return super().as_form_widget(**{**form_widget_options, "attrs": attrs})

    def _get_validators(self) -> Tuple[BaseValidator, ...]:
        """Return validators for handling past and future date limitations."""
        validators = cast(Tuple[BaseValidator, ...], ())

        if not self.allow_past:
            validators = (*validators, MinValueValidator(limit_value=datetime.now))

        if not self.allow_future:
            validators = (*validators, MaxValueValidator(limit_value=datetime.now))

        return validators


class DurationField(FieldType):
    """A field for collecting duration data."""

    label = "Duration"

    form_field_class = form_fields.DurationField
    model_field_class = model_fields.DurationField


class CheckboxField(FieldType):
    """A field for collecting a boolean value with a checkbox."""

    label = "Checkbox"

    form_field_class = form_fields.BooleanField
    form_widget_class = form_widgets.CheckboxInput
    form_widget_options = {"attrs": {"value": "true"}}
    model_field_class = model_fields.BooleanField


class YesNoRadioField(FieldType):
    """A field for collecting a yes/no value with radio buttons."""

    label = "Yes/No Radio Buttons"

    form_field_class = form_fields.TypedChoiceField
    form_field_options = {
        "choices": (
            ("True", "Yes"),
            ("False", "No"),
        ),
        "coerce": lambda v: True if v in ("True", True) else False,
    }
    form_widget_class = form_widgets.RadioSelect
    model_field_class = model_fields.BooleanField


class YesNoUnknownRadioField(FieldType):
    """A field for collecting a yes/no/unknown value with radio buttons."""

    label = "Yes/No/Unknown Radio Buttons"

    form_field_class = form_fields.TypedChoiceField
    form_field_options = {
        "choices": (
            ("True", "Yes"),
            ("False", "No"),
            ("None", "Unknown"),
        ),
        "coerce": lambda v: (
            True if v in ("True", True) else False if v in ("False", False) else None
        ),
    }
    form_widget_class = form_widgets.RadioSelect
    model_field_class = model_fields.BooleanField
    model_field_options = {"null": True}


class YesNoSelectField(FieldType):
    """A field for collecting a boolean value with a yes/no select field."""

    label = "Yes/No Dropdown"

    form_field_class = form_fields.TypedChoiceField
    form_field_options = {
        "choices": (
            ("True", "Yes"),
            ("False", "No"),
        ),
        "coerce": lambda v: True if v in ("True", True) else False,
    }
    form_widget_class = form_widgets.Select
    model_field_class = model_fields.BooleanField


class YesNoUnknownSelectField(FieldType):
    """A field for collecting a boolean value with a yes/no select field."""

    label = "Yes/No/Unknown Dropdown"

    form_field_class = form_fields.TypedChoiceField
    form_field_options = {
        "choices": (
            ("True", "Yes"),
            ("False", "No"),
            ("None", "Unknown"),
        ),
        "coerce": lambda v: (
            True if v in ("True", True) else False if v in ("False", False) else None
        ),
    }
    model_field_class = model_fields.BooleanField
    model_field_options = {"null": True}


class SingleChoiceSelectField(FieldType):
    """A field for collecting a single text value from a select list."""

    label = "Single-choice Dropdown"

    form_field_class = form_fields.ChoiceField
    form_field_options = {"choices": _EMPTY_CHOICES}
    model_field_class = model_fields.TextField


class SingleChoiceRadioSelectField(FieldType):
    """A field for collecting a text value from a set of radio buttons."""

    label = "Radio Buttons"

    form_field_class = form_fields.ChoiceField
    form_field_options = {"choices": _EMPTY_CHOICES}
    form_widget_class = form_widgets.RadioSelect
    model_field_class = model_fields.TextField


class MultipleChoiceSelectField(FieldType):
    """A field for collecting multiple text values from a select list."""

    label = "Multiple-choice Dropdown"

    form_field_class = form_fields.MultipleChoiceField
    form_field_options = {"choices": _EMPTY_CHOICES}
    model_field_class = JSONField


class MultipleChoiceCheckboxField(FieldType):
    """A field for collecting multiple text values from a checkbox list."""

    label = "Multiple-choice Checkboxes"

    form_field_class = form_fields.MultipleChoiceField
    form_field_options = {"choices": _EMPTY_CHOICES}
    form_widget_class = form_widgets.CheckboxSelectMultiple
    model_field_class = JSONField


class FileUploadField(FieldType):
    """A field for collecting file uploads."""

    label = "File Upload"

    form_field_class = form_fields.FileField
    model_field_class = FileField


class ImageUploadField(FieldType):
    """A field for collecting image uploads."""

    label = "Image Upload"

    form_field_class = form_fields.ImageField
    model_field_class = ImageField


class HiddenJSONField(FieldType):
    """A field for storing JSON data behind the scenes."""

    label = "Hidden JSON Field"

    form_field_class = JSONFormField
    form_widget_class = form_widgets.HiddenInput
    model_field_class = JSONField


AutocompleteResult = TypedDict("AutocompleteResult", {"id": str, "text": str})
AutocompleteResultMapping = TypedDict(
    "AutocompleteResultMapping",
    {"root": str, "value": str, "text": str, "extra": Dict[str, str]},
    total=False,
)


class AutocompleteJSONField(JSONFormField):
    """A Django form field for storing autocomplete results."""

    EMPTY_VALUES: Tuple[Any, ...] = (None, "null", "[]", [])

    def has_changed(self, initial: Any, data: Any) -> bool:
        """Returns True if the field value has changed.

        Considers empty lists and "null" to be equivalent to None.

        Args:
            initial: The initial value.
            data: The new value.

        Returns:
            bool: True if the field value has changed.
        """
        data = None if data in self.EMPTY_VALUES else data
        return cast(bool, super().has_changed(initial, data))


class BaseAutocompleteSelectField(FieldType):
    """An interface for implementing autocomplete select fields."""

    class Meta:
        abstract = True

    form_field_class = AutocompleteJSONField
    form_widget_class = AutocompleteSelect
    model_field_class = JSONField

    # Autocomplete-specific field_type_options.
    DEFAULT_MAPPING: AutocompleteResultMapping = {
        "root": "@",
        "value": "id",
        "text": "text",
        "extra": {},
    }
    mapping: AutocompleteResultMapping

    def as_form_widget(self, **form_widget_options: Any) -> form_widgets.Widget:
        """Build the autocomplete form widget.

        Replaces the url parameter with a proxy URL to a custom view to allow
        developers to implement fine-grained search behavior.

        Args:
            form_widget_options: Parameters passed to the form widget
                constructor.

        Returns:
            form_widgets.Widget: A configured widget for rendering the
                autocomplete input.
        """
        proxy_url = urlparse.urlparse(
            reverse(
                "flexible_forms:autocomplete",
                kwargs={
                    "app_label": self.field._meta.app_label,
                    "model_name": self.field._meta.model_name,
                    "field_pk": self.field.pk,
                },
            )
        )

        # Parse the query string so we can manipulate it.
        query_params = dict(urlparse.parse_qsl(proxy_url.query))

        # If a specific record instance was supplied, append its ID to the URL
        # so that autocomplete suggestions can be tailored to a particular
        # record's field values.
        if self.record and self.record.pk:
            query_params["record_pk"] = self.record.pk

        # Re-encode the querystring and replace the original one.
        proxy_url = proxy_url._replace(query=urlencode(query_params))

        return super().as_form_widget(
            **{
                **form_widget_options,
                "url": urlparse.urlunparse(proxy_url),
            },
        )

    def autocomplete(
        self, request: HttpRequest
    ) -> Tuple[Iterable[AutocompleteResult], bool]:
        """Return paginated search results for the given search_term.

        Returns an iterable of AutocompleteSearchResult objects and a boolean
        indicating whether there are more to be fetched via pagination.

        Args:
            request: The HTTP request.

        Returns:
            Tuple[Sequence[AutocompleteResult], bool]: A two-tuple containing
                a sequence of zero or more select2-compatible search results, and
                a boolean indicating whether or not there are more results to
                fetch with pagination.
        """
        # Extract GET parameters common to all autocomplete fields.
        search_term = request.GET.get("term", "")
        page = int(request.GET.get("page", "1"))
        per_page = int(request.GET.get("per_page", "50"))

        record = cast("BaseRecord", self.record)
        record_variable = (
            (record._meta.verbose_name or "record").lower().replace(" ", "_")
        )

        # Iterpolate the form_field_options using context from the current request.
        field_type_options = interpolate(
            self.field.field_type_options,
            context={
                "term": search_term,
                "page": page,
                "per_page": per_page,
                "record": record,
                record_variable: record,
                **record._data,
                **self.field_values,
            },
        )

        return self.get_results(
            request,
            search_term=search_term,
            page=page,
            per_page=per_page,
            **field_type_options,
        )

    def get_results(
        self,
        request: HttpRequest,
        search_term: str,
        page: int,
        per_page: int,
        **field_type_options: Any,
    ) -> Tuple[Iterable[AutocompleteResult], bool]:
        """Perform a search and return paginated results.

        Should return an iterable of autocomplete-compatible results and a
        boolean indicating whether there are more results to be fetched via
        pagination.

        Args:
            request: The HTTP request.
            search_term: The term to search for.
            page: The pagination page number.
            per_page: The pagination page size.
            field_type_options: The field's field_type_options after having
                its values interpolated.

        Raises:
            NotImplementedError: If the implementing class doesn't implement
                a get_results() method.
        """
        raise NotImplementedError(  # pragma: no cover
            f"All AutocompleteField types must implement get_results()."
        )


class URLAutocompleteSelectField(BaseAutocompleteSelectField):
    """A field for autocompleting selections from an HTTP endpoint."""

    label = "Autocomplete (URL)"

    # URL autocomplete-specific field type options.
    url: str
    search_fields: List[str]

    def get_results(
        self,
        request: HttpRequest,
        search_term: str,
        page: int,
        per_page: int,
        **field_type_options: Any,
    ) -> Tuple[Iterable[AutocompleteResult], bool]:
        """Perform a search and return paginated results.

        Search is performed via a GET request to the configured URL.

        If the configured URL is path only (e.g., "/search?query={{term}}")
        and that path can be resolved to a view function in the application,
        the view function will be called directly with the current request
        object (modified with the new path and GET parameters).

        Args:
            request: The HTTP request.
            search_term: The term to search for.
            page: The pagination page number.
            per_page: The pagination page size.
            field_type_options: The field's field_type_options after having
                its values interpolated.

        Returns:
            Tuple[Iterable[AutocompleteResult], bool]: An iterable of
                AutocompleteResult dicts, and a boolean indicating whether
                there are more results to be fetched via pagination.
        """
        results = cast(List[AutocompleteResult], [])
        has_more = False

        url: RenderedString = field_type_options["url"]

        if not url:
            logger.warning(
                f"The URL for field '{self.field}' was blank, and so no "
                f"autocomplete search could take place. This may be a "
                f"misconfiguration."
            )
            return results, has_more

        mapping = cast(
            AutocompleteResultMapping,
            {**self.DEFAULT_MAPPING, **field_type_options.get("mapping", {})},
        )
        search_manually: bool = "term" not in url.__context__

        # Get the JSON response from the endpoint.
        response_json = self._get_json_response(request, url)

        # Parse the search response and map each result to a Select2-compatible
        # dict with an "id" and a "text" property.
        raw_results = jp(mapping["root"], response_json, [])
        for raw_result in raw_results:
            result_text = str(jp(mapping["text"], raw_result))

            # If we're configured to search manually, skip results that don't
            # have the search term in the extracted text.
            if search_manually and search_term:
                if search_term.casefold() not in result_text.casefold():
                    continue

            results.append(
                create_autocomplete_result(
                    text=result_text,
                    value=jp(mapping["value"], raw_result),
                    extra={k: jp(v, raw_result) for k, v in mapping["extra"].items()},
                )
            )

        # If we're configured to paginate manually, then paginate.
        paginator = Paginator(results, per_page)
        paginated_page = paginator.page(page)
        results = cast(List[AutocompleteResult], paginated_page.object_list)
        has_more = paginated_page.has_next()

        return results, has_more

    def _get_json_response(self, request: HttpRequest, url: str) -> Any:
        """Make a request to the given URL and return the response JSON.

        Args:
            request: The current HttpRequest
            url: The URL to make a request to.

        Returns:
            Any: The JSON response body.
        """
        parsed_url = urlparse.urlparse(url)
        response_json = None

        # If the configured URL is just a path, we might be able to simply
        # proxy the request to an internal view function instead of making an
        # entirely new HTTP request. This is less expensive and side-steps
        # issues with auth when calling local views.
        if not parsed_url.netloc:
            try:
                resolver_match = resolve(parsed_url.path)
            except Resolver404:  # pragma: no cover
                pass
            else:
                view_func, args, kwargs = resolver_match

                # Invalidate cached properties affected by the request update.
                for attr, value in type(request).__dict__.items():
                    if not isinstance(value, cached_property):
                        continue
                    try:
                        delattr(request, attr)
                    except AttributeError:
                        pass

                # Update the request to point to the configured URL.
                request.META["QUERY_STRING"] = parsed_url.query
                request.GET = QueryDict(parsed_url.query)
                request.path = parsed_url.path
                request.path_info = parsed_url.path
                request.resolver_match = resolver_match

                # Call the view function with the updated request object.
                response = view_func(request, *args, **kwargs)
                response_json = json.loads(
                    response.rendered_content
                    if hasattr(response, "rendered_content")
                    else response.content
                )

        # If we weren't able to call a local view to get our results, we'll
        # make a regular GET request to the URL instead.
        if response_json is None:
            response = requests.get(request.build_absolute_uri(url))
            response.raise_for_status()
            response_json = response.json()

        return response_json


class URLAutocompleteSelectMultipleField(URLAutocompleteSelectField):
    """An autocomplete field that supports multiple selections."""

    label = "Autocomplete Multiple (URL)"
    form_widget_class = AutocompleteSelectMultiple


class QuerysetAutocompleteSelectField(BaseAutocompleteSelectField):
    """An autocomplete field that sources options from a queryset."""

    label = "Autocomplete (QuerySet)"

    model: str
    search_fields: List[str]
    filter: Dict[str, Any]
    exclude: Dict[str, Any]

    def get_results(
        self,
        request: HttpRequest,
        search_term: str,
        page: int,
        per_page: int,
        **field_type_options: Any,
    ) -> Tuple[Iterable[AutocompleteResult], bool]:
        """Perform a search and return paginated results.

        Search is performed with a QuerySet against the configured model
        using basic filters, along with any full-text search capabilities of
        the databse if available. Each model instance is then JSONified so
        that it can be queries with JMESPath by the configured mapping.

        Args:
            request: The HTTP request.
            search_term: The term to search for.
            page: The pagination page number.
            per_page: The pagination page size.
            field_type_options: The field's field_type_options after having
                its values interpolated.

        Returns:
            Tuple[Iterable[AutocompleteResult], bool]: An iterable of
                AutocompleteResult dicts, and a boolean indicating whether
                there are more results to be fetched via pagination.

        Raises:
            ImproperlyConfigured: If the configured search_field is not a
                field on the configured model (i.e. the field doesn't exist or
                is a property method, etc).
        """
        results = cast(List[AutocompleteResult], [])
        has_more = False

        # Extract the relevant configuration from the form widget options.
        model_name: str = field_type_options["model"]
        filter: Dict[str, Any] = field_type_options.get("filter", {})
        exclude: Dict[str, Any] = field_type_options.get("exclude", {})
        mapping = cast(
            AutocompleteResultMapping,
            {**self.DEFAULT_MAPPING, **field_type_options.get("mapping", {})},
        )

        # Build a set of model field names used in the mapping values. This
        # will be used to transform the database records into JSON before
        # applying the mapping.
        expression_fields: Set[str] = set()
        for expr in (*mapping.values(), *mapping["extra"].values()):
            if not isinstance(expr, str):
                continue
            expression_fields.update(get_expression_fields(expr))

        # Resolve the model class from the Django registry and build a queryset
        # with the given parameters.
        model_cls = cast(Type[DjangoModel], apps.get_model(model_name))
        model_fields = frozenset(
            [*(f.name for f in model_cls._meta.get_fields()), "pk"]
        )
        qs = (
            model_cls._default_manager.filter(**(filter or {}))
            .exclude(**(exclude or {}))
            .order_by("pk")
        )

        search_fields = field_type_options.get(
            "search_fields",
            # If no search_fields were configured explicitly, derive them from the
            # fields used in the "text" expression of the configured mapping.
            tuple(
                field_name
                for field_name in get_expression_fields(mapping["text"])
                if field_name in model_fields
            ),
        )

        # If any of the search_fields is not a model field, it can't be
        # used to filter results.
        if any(f.split("__")[0] not in model_fields for f in search_fields):
            raise ImproperlyConfigured(
                f"The search_fields option for {self.name} fields "
                f"must contain only the names of model fields."
            )

        # If no fields were specified to search, log a warning.
        if not search_fields:  # pragma: no cover
            logger.warning(
                f"No search_fields were defined or discovered, and so the "
                f"{self.name} field cannot filter results with a search "
                f"term."
            )

        # If a serch term is given, search for it using case-insensitive
        # queryset filters against any configured search_fields.
        if search_term and search_fields:
            # Look for a vendor-specific method for search, or use the fallback.
            search_method = getattr(
                self, f"_search_{connections[qs.db].vendor}", self._search_fallback
            )
            qs = search_method(qs, search_fields=search_fields, search_term=search_term)

        # Return distinct records (in case filters cause duplicates).
        qs = qs.distinct()

        # Paginate the queryset before mapping, since we don't know how big the
        # queryset could be.
        paginator = Paginator(qs, per_page)
        paginated_page = paginator.page(page)
        raw_results = paginated_page.object_list
        has_more = paginated_page.has_next()

        # Parse the search response and map each result to a Select2-compatible
        # dict with an "id" and a "text" property.
        for instance in raw_results:
            results.append(self.result_from_instance(instance, mapping=mapping))

        return results, has_more

    def result_from_instance(
        self, instance: DjangoModel, mapping: AutocompleteResultMapping
    ) -> AutocompleteResult:
        """Create an autocomplete-compatible result from a model instance.

        Uses the given mapping to generate the result data structure.

        Args:
            instance: The model instance to map to the result.
            mapping: The mapping to follow.

        Returns:
            AutocompleteResult: An autocomplete-compatible result.
        """
        expression_fields: Set[str] = set()
        for expr in (*mapping.values(), *mapping["extra"].values()):
            if not isinstance(expr, str):
                continue
            expression_fields.update(get_expression_fields(expr))

        instance_json = {}
        for expression_field in expression_fields:
            value = getattr(instance, expression_field, None)
            instance_json[expression_field] = str(value() if callable(value) else value)

        return create_autocomplete_result(
            text=str(jp(mapping["text"], instance_json)),
            value=jp(mapping["value"], instance_json),
            extra={k: jp(v, instance_json) for k, v in mapping["extra"].items()},
        )

    def _search_postgresql(
        self,
        queryset: "QuerySet[DjangoModel]",
        search_fields: Tuple[str, ...],
        search_term: str,
    ) -> "QuerySet[DjangoModel]":
        """Perform a search with a postgres backend.

        Uses postgresql-specific search strategies for better search results.

        Args:
            queryset: The queryset to filter with the given search parameters.
            search_fields: The names of the model fields to search.
            search_term: The term to search for.

        Returns:
            QuerySet[DjangoModel]: The filtered QuerySet.
        """
        from django.contrib.postgres.search import (
            SearchQuery,
            SearchRank,
            TrigramSimilarity,
        )

        # Check for trigram support (the pg_trgm PostgreSQL extension).
        db_connection = connections[queryset.db]
        supports_pg_trgm = cache.get_or_set(
            f"flexible_forms:db:{queryset.db}:supports_pg_trgm",
            lambda: check_supports_pg_trgm(db_connection),
        )

        # Annotate the queryset with search scores.
        search_rank_fields: Set[F] = set()
        for field_name in search_fields:
            strict_score_field = F(f"_{field_name}_strict_score")
            similarity_score_field = F(f"_{field_name}_similarity_score")
            search_rank_fields.update([strict_score_field, similarity_score_field])

            # Annotate the queryset with "strict" and "similarity" scores for
            # each field.
            #
            # The "strict" score is how well the search term matches the field
            # value with simple comparisons (iexact, istartswith, and
            # icontains).
            #
            # The "similarity" score uses PostgreSQL full-text search
            # capabilities (trigram similarity if available, otherwise tsvector
            # rankings with prefix search).
            queryset = queryset.annotate(
                **{
                    strict_score_field.name: Case(
                        When(**{f"{field_name}__iexact": search_term, "then": 3}),
                        When(**{f"{field_name}__istartswith": search_term, "then": 2}),
                        When(**{f"{field_name}__icontains": search_term, "then": 1}),
                        output_field=model_fields.IntegerField(),
                        default=0,
                    ),
                    # Annotate the queryset with a similarity score in addition
                    # to the strict matching.
                    similarity_score_field.name: (
                        # Use trigram similarity (pg_trgm) if it's supported.
                        TrigramSimilarity(field_name, search_term)
                        if supports_pg_trgm
                        else
                        # Use SearchRank (to_tsvector) as a fallback.
                        SearchRank(
                            vector=field_name,
                            # Split the search term into words, append a prefix
                            # operator to each one, and IOR them together for
                            # a sort of pseudo-lexeme search ranking.
                            query=reduce(
                                ior,
                                (
                                    SearchQuery(f"{word}:*", search_type="raw")
                                    for word in search_term.split(" ")
                                    if word
                                ),
                            ),
                        )
                    ),
                }
            )

        # Sum up all of the score fields into a single search rank.
        search_rank = Sum(
            reduce(add, search_rank_fields), output_field=model_fields.FloatField()
        )

        # Sum up all of the search scores into an aggregated search rank.
        queryset = queryset.annotate(_search_rank=search_rank)
        # Filter out any results that are total non-matches.
        queryset = queryset.filter(_search_rank__gt=0.0)
        # Order the results by the search rank, highest first.
        queryset = queryset.order_by(f"-_search_rank")

        return queryset

    def _search_fallback(
        self,
        queryset: "QuerySet[DjangoModel]",
        search_fields: Tuple[str, ...],
        search_term: str,
    ) -> "QuerySet[DjangoModel]":
        """Perform a search with a generic backend.

        Args:
            queryset: The queryset to filter with the given search parameters.
            search_fields: The names of the model fields to search.
            search_term: The term to search for.

        Returns:
            QuerySet[DjangoModel]: The filtered QuerySet.
        """
        # Annotate the queryset with search scores.
        search_rank_fields: Set[F] = set()
        for field_name in search_fields:
            strict_score_field = F(f"_{field_name}_strict_score")
            search_rank_fields.add(strict_score_field)

            # Annotate the queryset with a search score based on how well the
            # field value matches the search term.
            queryset = queryset.annotate(
                **{
                    strict_score_field.name: Case(
                        When(**{f"{field_name}__iexact": search_term, "then": 3}),
                        When(**{f"{field_name}__istartswith": search_term, "then": 2}),
                        When(**{f"{field_name}__icontains": search_term, "then": 1}),
                        output_field=model_fields.IntegerField(),
                        default=0,
                    ),
                }
            )

        # Sum up all of the score fields into a single search rank.
        search_rank = Sum(
            reduce(add, search_rank_fields), output_field=model_fields.FloatField()
        )

        # Sum up all of the search scores into an aggregated search rank.
        queryset = queryset.annotate(_search_rank=search_rank)
        # Filter out any results that are total non-matches.
        queryset = queryset.filter(_search_rank__gt=0)
        # Order the results by the search rank, highest first.
        queryset = queryset.order_by(f"-_search_rank")

        return queryset

    def apply_value(
        self,
        form_field: form_fields.Field,
        value: Any,
        **kwargs: Any,
    ) -> form_fields.Field:
        """Set the value of the autocomplete field.

        Given a model instance or primary key as the value, returns a mapped
        autocomplete-compatible result.

        Args:
            form_field: The field for which to set the value.
            value: The model instance or primary key for which to generate an
                autocomplete result.
            kwargs: Unused.

        Returns:
            form_fields.Field: The given form field, modified to include the
                new value.
        """
        model_instance = value if isinstance(value, DjangoModel) else None

        # If the given value isn't a model instance, assume it's the primary key of the instance and look it up.
        if not model_instance:
            model_cls = cast(Type[DjangoModel], apps.get_model(self.model))
            model_instance = model_cls._default_manager.get(pk=value)

        # Map the retrieved model instance to an autocomplete-friendly result dict.
        result = self.result_from_instance(
            model_instance,
            mapping=cast(
                AutocompleteResultMapping,
                {**self.DEFAULT_MAPPING, **self.mapping},
            ),
        )

        return super().apply_value(form_field, result, **kwargs)


class QuerysetAutocompleteSelectMultipleField(QuerysetAutocompleteSelectField):
    """A queryset autocomplete field that supports multiple selections."""

    label = "Autocomplete Multiple (QuerySet)"
    form_widget_class = AutocompleteSelectMultiple



# -*- coding: utf-8 -*-

"""Common utilities."""

import json
from functools import lru_cache, singledispatch
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import jmespath
from django.db import DatabaseError
from django.db.backends.base.base import BaseDatabaseWrapper
from django.template import Context, Template
from django.template.base import VariableNode
from jmespath.parser import Parser
from simpleeval import (
    DEFAULT_FUNCTIONS,
    DEFAULT_OPERATORS,
    EvalWithCompoundTypes,
)

if TYPE_CHECKING:  # pragma: no cover
    from flexible_forms.fields import AutocompleteResult

T = TypeVar("T", bound=Type)


def empty(value: Any) -> bool:
    """Return True if the given value is "empty".

    Considers a value empty if it's iterable and has no elements, of if the
    value is in the given sequence of `empty_values` (only None by default).

    Args:
        value: The value to test for emptiness.

    Returns:
        bool: True if the given value is empty.
    """
    if hasattr(value, "__iter__"):
        try:
            next(iter(value))
        except StopIteration:
            return True
        return False

    return value is None


class FormEvaluator(EvalWithCompoundTypes):
    """An evaluator subclass for evaluating form expressions."""

    OPERATORS = {
        **DEFAULT_OPERATORS.copy(),
    }

    FUNCTIONS = {
        **DEFAULT_FUNCTIONS.copy(),
        "empty": empty,
    }

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs = {
            **kwargs,
            "operators": self.OPERATORS,
            "functions": self.FUNCTIONS,
        }
        super().__init__(*args, **kwargs)


def evaluate_expression(
    expression: str,
    names: Optional[
        Mapping[
            str,
            Any,
        ]
    ] = None,
    **kwargs: Any,
) -> Any:
    """Safely evaluate a Python expression.

    Evaluates a Python expression in a controlled environment.

    Args:
        expression: The Python expression to evaluate.
        names: A mapping of variable names and
            their values available to the expression.
        kwargs: Passed to the FormEvaluator constructor.

    Returns:
        Any: The value of the expression, cast using the given `cast`
            callable if specified.
    """
    evaluator = FormEvaluator(names=names, **kwargs)
    return evaluator.eval(expression)


def replace_element(
    needle: Any,
    replacement: Any,
    haystack: Union[List[Any], Tuple[Any, ...]],
) -> Union[List[Any], Tuple[Any, ...]]:
    """Replace a value recursively in a given data structure.

    Args:
        needle: The element to replace.
        replacement: The replacement value.
        haystack: The data structure in which to find and replace values.

    Returns:
        Union[List[Any], Tuple[Any, ...]]: A new data structure of the given
            type with the desired elements replaced.
    """
    elements = type(haystack)()
    for element in haystack:
        if isinstance(element, str):
            element = replacement if element == needle else element
        else:
            element = replace_element(needle, replacement, element)
        elements = type(haystack)([*elements, element])
    return elements


def stable_json(data: Union[dict, list, None]) -> str:
    """Generate a stable string representation of the given dict or list.

    Args:
        data: The dict or list for which to produce a stable JSON
            representation.

    Returns:
        str: A stable JSON string representation of the given data.
    """
    return json.dumps(
        data, sort_keys=True, ensure_ascii=True, separators=(",", ":"), default=str
    )


def jp(expr: str, data: Any, default: Any = None) -> Any:
    """A shorthand helper for querying dicts with jmespath.

    Args:
        expr: The JMESPath expression.
        data: The data to query.
        default: The default value to return if the query returns None.

    Returns:
        Any: The result of the query, or the value of default.
    """
    result = jmespath.search(expression=expr, data=data)
    return default if result is None else result


@lru_cache(128)
def get_expression_fields(jmespath_expression: str) -> Tuple[str, ...]:
    """Return a list of fields referenced in the given JMESPath expression.

    Args:
        jmespath_expression: The expression for which to extract a list of
            referenced fields.

    Returns:
        Set[str]: A set containing the names of fields referenced in the
            expression.
    """
    return tuple(_get_fields(Parser().parse(jmespath_expression).parsed).keys())


def _get_fields(
    node: Dict[str, Any], ignore_fields: Sequence[str] = ("null",)
) -> Dict[str, None]:
    referenced_fields: Dict[str, None] = {}

    node_type = node["type"]
    children = node["children"]
    value = node.get("value")

    if node_type == "field" and value not in ignore_fields:
        referenced_fields[str(value)] = None
    elif node_type == "subexpression":
        referenced_fields.update(_get_fields(children[0]))
    else:
        for child in children:
            referenced_fields.update(_get_fields(child))

    return referenced_fields


class NOT_PROVIDED:
    """A proxy type for specifying that something was not defined.

    Useful when None is a valid value.
    """


class RenderedString(str):
    """A string with an attribute containing the render context.

    The __context__ property will be a dict containing the keys of the
    variables used to render the string. Variables in the original
    context that were not used to render the string will be excluded.
    Variables that were referenced in the string but not provided in the
    context should use the NOT_PROVIDED proxy value.
    """

    __context__: Dict[str, Any]


@singledispatch
def interpolate(data: Any, context: Dict[str, Any], strict: bool = True) -> Any:
    """Interpolate the given data using DTL.

    Handles (nested) dicts and lists and will interpolate strings containing DTL tokens.

    Args:
        data: The data to be interpolated.
        context: The context with which to interpolate strings containing DTL tokens.
        strict: If true, throws an error when the data references a variable
            that is not available in the context.

    Returns:
        Any: The interpolated data.
    """
    # The default implementation returns the data as-is.
    return data


@interpolate.register(str)
def _interpolate_str(
    data: str, context: Dict[str, Any], strict: bool = True
) -> RenderedString:
    """Handles interpolation of string values.

    Interpolates strings using the default Django Template Language engine.

    Args:
        data: The data to be interpolated.
        context: The context with which to interpolate strings containing DTL tokens.
        strict: If True, throws an error when the data references a variable
            that is not available in the context.

    Raises:
        LookupError: if strict is True and the data references a variable that
            is not available in the context.

    Returns:
        RenderedString: The rendered string with a __context__ property
            containing the variables used in rendering.
    """
    # Render the given string as a Django template with the given context.
    template = Template(data)
    template_context = Context(context, autoescape=False)
    rendered_string = RenderedString(template.render(template_context))

    # Extract a dict of variables used to render the string.
    rendered_context = {
        var: template_context.get(var, NOT_PROVIDED)
        for var in frozenset(
            v.filter_expression.var.lookups[0]
            for v in template.nodelist
            if isinstance(v, VariableNode)
        )
    }

    # Attach the render context to the string.
    rendered_string.__context__ = rendered_context

    missing_variables = frozenset(
        k for k, v in rendered_context.items() if v is NOT_PROVIDED
    )
    if strict and missing_variables:
        raise LookupError(
            f"The template references variables that were not in the context "
            f'provided: {", ".join(missing_variables)}'
        )

    return rendered_string


@interpolate.register(dict)
def _interpolate_dict(data: dict, context: Dict[str, Any], strict: bool = True) -> dict:
    """Handles interpolation of dict values.

    Interpolates the values of the dict as appropriate (renders strings, or
    recurses for list or dict values).

    Args:
        data: The data to be interpolated.
        context: The context with which to interpolate strings containing DTL tokens.
        strict: If True, throws an error when the data references a variable
            that is not available in the context.

    Returns:
        dict: The interpolated dict.
    """
    return {k: interpolate(v, context, strict) for k, v in data.items()}


@interpolate.register(list)
def _interpolate_list(data: list, context: Dict[str, Any], strict: bool = True) -> list:
    """Handles interpolation of list values.

    Interpolates the elements of the list as appropriate (renders strings, or
    recurses for list or dict elements).

    Args:
        data: The list to be interpolated.
        context: The context with which to interpolate strings containing DTL tokens.
        strict: If True, throws an error when the data references a variable
            that is not available in the context.

    Returns:
        list: The interpolated list.
    """
    return [interpolate(v, context, strict) for v in data]


# Determine if the database support trigram similarity by checking for
# the pg_trgm extension.
def check_supports_pg_trgm(connection: BaseDatabaseWrapper) -> bool:
    """Determine if trigram similarity support is present.

    Attempts to run a query against the extensions table in PostgreSQL to
    look for the pg_trgm extension. Failures assume a lack of support.

    Args:
        connection: The database connection.

    Returns:
        bool: True if trigram support is present.
    """
    try:
        with connection.cursor() as cursor:
            cursor.execute(
                """
                SELECT 1
                FROM pg_extension
                WHERE extname = 'pg_trgm' LIMIT 1;
            """
            )
            return bool(cursor.fetchone())
    except DatabaseError:
        return False


def collect_annotations(cls: Type) -> Dict[str, Type]:
    """Collects annotations from an object hierarchy.

    Args:
        cls: The class from which to extract property annotations.

    Returns:
        dict: A dict of each annotation's name and type.
    """
    annotations = cast(Dict[str, Type], getattr(cls, "__annotations__", {}))
    for base in cls.__bases__:
        annotations = {**annotations, **collect_annotations(base)}
    return annotations


def create_autocomplete_result(
    text: str,
    value: Any,
    extra: Optional[Dict[str, Any]] = None,
) -> "AutocompleteResult":
    """Generate a valid value for an AutocompleteSelect field.

    The AutocompleteSelect field types require options to be structured in a
    specific way in order to optimize subsequent renders: the entire
    autocomplete result is serialized as JSON in the ID field to prevent
    unnecessary lookups when rendering an existing value into a form.

    Args:
        text: The displayed text of the autocomplete option.
        value: The unique value assigned to the option.
        extra: Additional data stored along with the option.

    Returns:
        AutocompleteResult: A valid autocomplete result dict.
    """
    result = {
        "text": text,
        "value": text if value is None else value,
        "extra": extra or {},
    }

    # The "id" is the value eventually stored in the database. In this
    # case, it is a stringified version of the result JSON so that it can
    # be deserialized when rendering the initial value for the widget.
    #
    # The use of "id" as opposed to "value" or something that makes
    # more semantic sense is to support the use of the select2
    # implementation that ships with the Django admin.
    result["id"] = stable_json(result)

    return cast("AutocompleteResult", result)


def create_autocomplete_value(*args: Any, **kwargs: Any) -> str:
    """Return a stabilized JSON string of an autocomplete result.

    Acts as a thin wrapper for create_autocomplete_result.

    Args:
        args: Args passed to create_autocomplete_result
        kwargs: Kwargs passed to create_autocomplete_result

    Returns:
        str: A stabilized JSON string of the autocomplete result.
    """
    return stable_json(
        cast(Dict[str, Any], create_autocomplete_result(*args, **kwargs))
    )



# -*- coding: utf-8 -*-
from django.dispatch import Signal

# Signals emitted before and after preparing a Django form class from a BaseForm.
pre_form_class_prepare = Signal()
post_form_class_prepare = Signal()

# Signals emitted before and after building the fieldsets for a BaseForm.
pre_fieldsets_prepare = Signal()
post_fieldsets_prepare = Signal()

# Signals emitted before and after initializing a Record form.
pre_form_init = Signal()
post_form_init = Signal()

# Signals emitted before and after cleaning and validating a Record form.
pre_form_clean = Signal()
post_form_clean = Signal()

# Signals emitted before and after saving a Record form.
pre_form_save = Signal()
post_form_save = Signal()



# -*- coding: utf-8 -*-

"""Model definitions for the flexible_forms module."""

import inspect
import re
import weakref
from itertools import groupby
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Match,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypedDict,
    TypeVar,
    Union,
    cast,
)

from django import forms
from django.core.exceptions import (
    ImproperlyConfigured,
    ObjectDoesNotExist,
    ValidationError,
)
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models, transaction
from django.db.models.fields.mixins import FieldCacheMixin
from django.db.models.fields.related import (  # type: ignore
    ForeignObject,
    lazy_related_operation,
)
from django.db.models.options import Options
from django.db.models.query import Prefetch
from django.db.models.signals import class_prepared, pre_init
from django.dispatch.dispatcher import receiver
from django.forms.widgets import Widget
from django.utils.functional import cached_property
from django.utils.text import slugify
from simpleeval import FunctionNotDefined, NameNotDefined

from flexible_forms.fields import FIELD_TYPES, FieldType
from flexible_forms.signals import (
    post_fieldsets_prepare,
    post_form_class_prepare,
    pre_fieldsets_prepare,
    pre_form_class_prepare,
)
from flexible_forms.utils import (
    FormEvaluator,
    evaluate_expression,
    replace_element,
)

try:
    from django.db.models import JSONField
except ImportError:  # pragma: no cover
    from django.contrib.postgres.fields import JSONField  # type: ignore

# If we're only type checking, import things that would otherwise cause an
# ImportError due to circular dependencies.
if TYPE_CHECKING:  # pragma: no cover
    from flexible_forms.forms import BaseRecordForm

M = TypeVar(
    "M",
    bound="FlexibleBaseModel",
)

##
# FIELD_TYPE_OPTIONS
#
# A choice-field-friendly list of all available field types. Sorted for
# migration stability.
#
FIELD_TYPE_OPTIONS = sorted(
    ((k, v.label) for k, v in FIELD_TYPES.items()),
    key=lambda o: o[0],
)


class ProxyDescriptor:
    """Proxy attribute access to another attribute."""

    original_name: str

    def __init__(self, original_name: str):
        """Establish a proxy descriptor.

        Args:
            original_name: The name of the attribute on an instance to
                which attribute access should be proxied.
        """
        self.original_name = original_name

    def __get__(self, instance: Optional[object], cls: Optional[Type] = None) -> Any:
        """Get an attribute's value.

        Proxies the __get__ call to the configured original_name on the
        instance (or the cls, if instance is None).

        Args:
            instance: The object from which to get the value.
            cls: The type of the instance object.

        Returns:
            Any: the value of the attribute.
        """
        return getattr(cls if instance is None else instance, self.original_name)

    def __set__(self, instance: object, value: Any) -> None:
        """Set an attribute's value.

        Proxies the __set__ call to the attribute with the configured
        original_name on the instance.

        Args:
            instance: The object onto which to set the value.
            value: The value.
        """
        setattr(instance, self.original_name, value)

    def __delete__(self, instance: object) -> None:
        """Delete an attribute's value.

        Proxies the __delete__ call to the attribute with the configured
        original_name on the instance.

        Args:
            instance: The object from which to delete the value.
        """
        delattr(instance, self.original_name)


class AliasField(models.Field):
    """A model field that defines a transparent alias to another field."""

    original_field: models.Field
    concrete: bool = False

    def __init__(self, original_field: models.Field, **kwargs: Any) -> None:
        # Hacky, but: changes its class bases to mimic those of the original
        # field.
        self.__class__.__bases__ = original_field.__class__.__bases__

        # Prevent an AppRegistryNotReady error when deconstructing the original
        # field.
        original_swappable = getattr(original_field, "swappable", None)
        setattr(original_field, "swappable", False)

        # Deconstruct the field so that we can use its attributes to build the
        # alias field.
        _, _, _, original_kwargs = original_field.deconstruct()

        # Restore the original_swappable value, if necessary.
        setattr(original_field, "swappable", original_swappable)

        # Initialize the field.
        super().__init__(**{**original_kwargs, **kwargs})

        self.original_field = original_field

    def contribute_to_class(
        self, cls: Type[models.Model], name: str, private_only: bool = True
    ) -> None:
        """Define the alias field and add it to the model class.

        Handles the model field definition and adds proxy descriptors for
        transparent field proxying.

        Args:
            cls: The model class on which the field is defined.
            name: The name of the class attribute to which the field is
                assigned.
            private_only: Whether the field should be considered "private"
                by the ORM or not. Usually false, but True for the
                AliasField to prevent it from interfering with migrations
                and other core Django facilities.
        """
        self.name = name
        self.attname = super().get_attname()
        self.column = self.original_field.column
        self.model = self.original_field.model

        # Add the field to the set of fields for the model.
        cls._meta.add_field(self, private=private_only)

        # Create proxies to the original descriptors.
        setattr(cls, self.name, ProxyDescriptor(self.original_field.name))
        setattr(cls, self.attname, ProxyDescriptor(self.original_field.attname))

        # Relation fields should have their reverse descriptors proxied on
        # their remote model classes.
        if self.is_relation:
            lazy_related_operation(
                lambda _, related, field: setattr(
                    related,
                    field.remote_field.get_accessor_name(),
                    ProxyDescriptor(
                        field.original_field.remote_field.get_accessor_name(),
                    ),
                ),
                cls,
                self.original_field.remote_field.model,
                field=self,
            )

        # Connect a model signal to sync
        self._connect_signals(cls)

    def get_cache_name(self) -> str:
        """Get the cache key for the field.

        Delegates this responsibility to the original field so that the
        original field and alias field can use the cache interchangeably.

        Returns:
            str: The cache key for the field.
        """
        return cast(FieldCacheMixin, self.original_field).get_cache_name()

    def _connect_signals(self, cls: Type[models.Model]) -> None:
        """Connect model signals for the given model class.

        Ensures that values between the alias and original fields are
        synchronized before attempting model operations.

        Args:
            cls: The class on which to listen for signals.
        """

        def _on_pre_init(
            *,
            kwargs: Any,
            original_field: models.Field = self.original_field,
            alias_field: models.Field = self,
            **_: Any,
        ) -> None:
            original_name = original_field.name
            original_attname = original_field.attname
            alias_name = alias_field.name
            alias_attname = alias_field.attname

            if original_name in kwargs and alias_name not in kwargs:
                kwargs[alias_name] = kwargs[original_name]
            if original_attname in kwargs and alias_attname not in kwargs:
                kwargs[alias_attname] = kwargs[original_attname]
            if alias_name in kwargs and original_name not in kwargs:
                kwargs[original_name] = kwargs[alias_name]
            if alias_attname in kwargs and original_attname not in kwargs:
                kwargs[original_attname] = kwargs[alias_attname]

        pre_init.connect(_on_pre_init, sender=cls, weak=False)


class FlexibleRelation(ForeignObject):
    """A custom field for defining a swappable ForeignKey field for a model."""

    field_class = Type[ForeignObject]

    to_base_model: Type["FlexibleBaseModel"]

    original_field_name: str
    original_related_name: Optional[str]
    original_args: Sequence[Any]
    original_kwargs: Dict[str, Any]

    def __init__(
        self,
        to: Type["FlexibleBaseModel"],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.to_base_model = to
        self.original_related_name = cast(Optional[str], kwargs.get("related_name"))
        self.original_args = args
        self.original_kwargs = kwargs.copy()

        super().__init__(to, *args, **kwargs)

    @property
    def target_field(self) -> models.Field:
        """Return the target field for the relation."""
        try:
            return super().target_field
        except IndexError:  # pragma: no cover
            return cast("models.Field", None)

    def contribute_to_class(
        self, cls: Type[models.Model], name: str, private_only: bool = False
    ) -> None:
        """Add the relation and optional alias fields to the class.

        Args:
            cls: The class to contribute to.
            name: The name of the attribute that the FlexibleRelation was assigned to.
            private_only: Unused.
        """
        cls = cast(Type[FlexibleBaseModel], cls)
        self.original_field_name = name

        if cls._meta.abstract or not hasattr(cls, "flexible_forms"):
            super().contribute_to_class(cls, name, private_only=private_only)
            return

        # Resolve the appropriate concrete model and field names for the configured _to_base_model.
        to_concrete_model = cls.flexible_forms.get_model(
            self.to_base_model, create=True
        )
        field_name = getattr(cls.FlexibleMeta, f"{name}_field_name")
        related_name = getattr(cls.FlexibleMeta, f"{name}_field_related_name")

        # Invoke the field constructor now that we've discovered all of the
        # necessary flexible configuration, then add it to the class with
        # contribute_to_class.
        super().__init__(
            to_concrete_model,
            *self.original_args,
            **{**self.original_kwargs, "related_name": related_name},
        )
        super().contribute_to_class(cls, field_name, private_only=private_only)

        # If the field name configured in the FlexibleMeta differs from the
        # original name, add an alias field so that this field can be accessed
        # transparently from either property.
        if field_name != self.original_field_name:
            alias_related_name = self.original_related_name
            if alias_related_name == related_name:
                alias_related_name = f"{alias_related_name}+"

            alias_field = AliasField(
                original_field=self,
                to=to_concrete_model,
                related_name=alias_related_name,
            )
            alias_field.contribute_to_class(
                cls, self.original_field_name, private_only=True
            )

            self._reconfigure_meta(cls._meta, self.original_field_name, field_name)

        # HACK: Change the base class to fool the migration dector to use the
        # appropriate Django relationship field type instead of this custom
        # field class.
        self.__class__ = self.field_class  # type: ignore

    def _reconfigure_meta(
        self, meta: Options, from_field: str, to_field: str
    ) -> Options:
        # Replace any references to old field names in order_with_respect_to.
        meta.order_with_respect_to = (
            to_field  # type: ignore
            if meta.order_with_respect_to == from_field
            else meta.order_with_respect_to
        )
        meta.original_attrs["order_with_respect_to"] = meta.order_with_respect_to

        # Replace any references to old field names in unique_together.
        meta.unique_together = replace_element(
            from_field, to_field, meta.unique_together
        )
        meta.original_attrs["unique_together"] = meta.unique_together

        # Replace any references to old field names in unique_together.
        meta.index_together = replace_element(from_field, to_field, meta.index_together)
        meta.original_attrs["index_together"] = meta.index_together

        # Replace any references to old field names in configured indexes.
        indexes = []
        for index in meta.indexes:
            index_class = index.__class__
            _, _, kwargs = index.deconstruct()
            kwargs["fields"] = replace_element(from_field, to_field, kwargs["fields"])
            indexes.append(index_class(**kwargs))
        meta.indexes = indexes
        meta.original_attrs["indexes"] = meta.indexes

        # Replace any references to old field names in configured constraints.
        constraints = []
        for constraint in meta.constraints:
            constraint_class = constraint.__class__
            _, _, kwargs = constraint.deconstruct()
            kwargs["fields"] = replace_element(from_field, to_field, kwargs["fields"])
            constraints.append(constraint_class(**kwargs))
        meta.constraints = constraints
        meta.original_attrs["constraints"] = meta.constraints

        return meta


class FlexibleForeignKey(FlexibleRelation, models.ForeignKey):
    """A flexible foreign key relation field."""

    field_class = models.ForeignKey


class FlexibleOneToOneField(FlexibleRelation, models.OneToOneField):
    """A flexible one-to-one relation field."""

    field_class = models.OneToOneField


class FlexibleBaseModel(models.Model):
    """A common base for all FlexibleField base models."""

    class Meta:
        abstract = True

    FlexibleMeta: Type[Any]

    flexible_forms: "FlexibleForms"

    @classmethod
    def _flexible_model_for(cls, base_model: Type[M]) -> Type[M]:
        """Return the current implementation of the given base_model.

        Args:
            base_model: The flexible_forms base model for which to return the implementation.

        Returns:
            Type[FleixbleModel]: The model class for the given base model in
                the current implementation.
        """
        return cls.flexible_forms.get_model(base_model)


@receiver(class_prepared)
def register_flexible_model(sender: Type[models.Model], **kwargs: Any) -> None:
    """Register a new flexible model.

    When a FlexibleBaseModel implementation has been prepared, it is
    registered with the appropriate FlexibleForms object so that other
    dependent FlexibleForms models can reference it.

    Args:
        sender: The prepared model class.
        kwargs: Unused.
    """
    if not issubclass(sender, FlexibleBaseModel) or sender._meta.abstract:
        return
    sender.flexible_forms.register_model(sender)


class _DjangoFieldsetOpts(TypedDict):
    """Required configuration for a Django fieldset."""

    fields: Sequence[Union[str, Sequence[str]]]


class DjangoFieldsetOpts(_DjangoFieldsetOpts, total=False):
    """Optional configuration for a Django fieldset."""

    description: Optional[str]
    classes: Sequence[str]


DjangoFieldset = Tuple[Optional[str], DjangoFieldsetOpts]


class BaseForm(FlexibleBaseModel):
    """A model representing a single type of customizable form."""

    label = models.TextField(
        blank=True,
        default="",
        help_text="The human-friendly name of the form.",
    )
    name = models.TextField(
        blank=True,
        help_text=(
            "The machine-friendly name of the form. Computed automatically "
            "from the label if not specified."
        ),
    )

    fields: "models.BaseManager[BaseField]"
    fieldsets: "models.BaseManager[BaseFieldset]"
    records: "models.BaseManager[BaseRecord]"

    class Meta:
        abstract = True

    def __str__(self) -> str:
        return self.label or f"Untitled {cast(str, self._meta.verbose_name).title()}"

    def save(self, *args: Any, **kwargs: Any) -> None:
        """Save the model.

        Sets the name if not specified.

        Args:
            args: (Passed to super)
            kwargs: (Passed to super)
        """
        self.name = self.name or slugify(self.label).replace("-", "_")

        super().save(*args, **kwargs)

    @property
    def initial_values(self) -> Dict[str, Any]:
        """Return a mapping of initial values for the form."""
        return {
            **{f.name: f.initial for f in self.fields.all()},
            "form": self,
        }

    def as_django_fieldsets(self) -> Sequence[DjangoFieldset]:
        """Generate a Django fieldsets configuration for the form.

        The Django admin supports the specification of fieldsets -- a
        simple way of grouping fields together. This property builds
        """
        django_fieldsets: Sequence[DjangoFieldset] = []
        fieldsets = self.fieldsets.all()

        pre_fieldsets_prepare.send(sender=self.__class__, instance=self)

        seen_fields = set()
        for fieldset in fieldsets:
            fieldset_items: Union[Tuple, Tuple[Union[Sequence[str], str]]] = ()

            # Sort the fieldset items by their vertical order, then group them
            # by that order. This creates "rows" of items that have the same
            # vertical_order.
            vertically_sorted_items = sorted(
                fieldset.items.all(), key=lambda f: int(f.vertical_order)
            )
            vertical_groups = groupby(
                vertically_sorted_items,
                lambda f: int(f.vertical_order),
            )

            # For each of the vertical groups that were created, sort them by
            # their horizontal_order. This sets their horizontal display order
            # so that items with a lower horizontal order are displayed first.
            for _order, vertical_group in vertical_groups:
                sorted_group = [
                    i.field.name
                    for i in sorted(
                        vertical_group, key=lambda i: int(i.horizontal_order)
                    )
                ]
                seen_fields.update(sorted_group)
                fieldset_items = (
                    *fieldset_items,
                    tuple(sorted_group) if len(sorted_group) > 1 else sorted_group[0],
                )

            # Add the configured fieldset to the rest of them.
            django_fieldsets = [
                *django_fieldsets,
                (
                    fieldset.name or None,
                    {
                        "classes": tuple(filter(bool, fieldset.classes.split(" "))),
                        "description": fieldset.description or None,
                        "fields": fieldset_items,
                    },
                ),
            ]

        post_fieldsets_prepare.send(
            sender=self.__class__,
            instance=self,
            fieldsets=django_fieldsets,
        )

        return django_fieldsets

    def as_django_form(
        self,
        data: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
        instance: Optional["BaseRecord"] = None,
        initial: Optional[Mapping[str, Any]] = None,
        exclude: Iterable[str] = (),
        **kwargs: Any,
    ) -> "BaseRecordForm":
        """Return the form represented as a Django form instance.

        Args:
            data: Data to be passed to the Django form constructor.
            files: Data to be passed to the Django form constructor.
            instance: The Record instance to be passed to the Django form
                constructor.
            initial: Initial values for the form fields.
            exclude: Field names to exclude from the form.
            **kwargs: Passed to the Django ModelForm constructor.

        Returns:
            BaseRecordForm: A configured BaseRecordForm (a Django ModelForm instance).
        """
        RecordModel = self._flexible_model_for(BaseRecord)
        form_field_name = RecordModel.FlexibleMeta.form_field_name

        # Exclude the form relationships. Calling this method implies that they
        # are already set, and they cannot be changed via normal means.
        if form_field_name != "form":
            exclude = (*exclude, "form")

        all_fields = tuple(f for f in self.fields.all() if f.name not in exclude)

        # Build a dict containing all field values. This combines all of the
        # form data into a single structure that will be used when evaluating
        # expressions against the form state.
        field_values: Dict[str, Any] = {
            **self.initial_values,
            **(instance._data if instance else {}),
            **(data or {}),
            **(files or {}),
        }

        # Regenerate the form fields, this time taking the field values into
        # account in order to inform any dynamic behaviors.
        form_fields = {
            f.name: f.as_form_field(field_values=field_values, record=instance)
            for f in all_fields
        }

        # Import the form class inline to prevent a circular import.
        from flexible_forms.forms import BaseRecordForm

        # Send a signal before preparing the form class to give receivers an
        # opportunity to modify the form structure.
        pre_form_class_prepare.send(
            sender=self.__class__,
            data=data,
            files=files,
            instance=instance,
            initial=initial,
            exclude=exclude,
            form_fields=form_fields,
        )

        form_name = f"{self.name.title().replace('_', '')}Form"
        form_class: Type[BaseRecordForm] = type(
            form_name,
            (BaseRecordForm,),
            {
                "__module__": self.__module__,
                "Meta": type(
                    "Meta",
                    (BaseRecordForm.Meta,),
                    {
                        "model": RecordModel,
                        "exclude": exclude,
                    },
                ),
                **form_fields,
            },
        )

        # Send a signal before preparing the form class to give receivers an
        # opportunity to modify the form class after creation.
        post_form_class_prepare.send(
            sender=self.__class__,
            data=data,
            files=files,
            instance=instance,
            initial=initial,
            exclude=exclude,
            form_fields=form_fields,
            form_class=form_class,
        )

        initial = {form_field_name: self, **(initial or {})}

        # Create a form instance from the form class and the passed parameters.
        form_instance = form_class(
            data=data,
            files=files,
            instance=instance,
            initial=initial,
            **kwargs,
        )

        return form_instance


class BaseField(FlexibleBaseModel):
    """A field on a form.

    A field belonging to a Form. Attempts to emulate a subset of
    Django's Field interface for that forms can be built dynamically.
    """

    label = models.TextField(
        help_text="The label to be displayed for this field in the form.",
    )

    name = models.TextField(
        blank=True,
        help_text=(
            "The machine-friendly name of the field. Computed automatically "
            "from the label if not specified."
        ),
    )

    label_suffix = models.TextField(
        blank=True,
        default="",
        help_text='The character(s) at the end of the field label (e.g. "?" or ":").',
    )

    help_text = models.TextField(
        blank=True,
        default="",
        help_text="Text to help the user fill out the field.",
    )

    required = models.BooleanField(
        default=False,
        help_text="If True, requires a value be present in the field.",
    )

    initial = JSONField(
        blank=True,
        null=True,
        help_text=("The default value if no value is given during initialization."),
        encoder=DjangoJSONEncoder,
    )

    error_messages = JSONField(
        blank=True,
        null=True,
        default=dict,
        help_text="Custom validation error messages.",
    )

    localize = models.BooleanField(
        blank=True,
        default=False,
        help_text="If True, enables localization support for this field.",
    )

    disabled = models.BooleanField(
        blank=True,
        default=False,
        help_text="If True, disables the field. Only really useful when used with a modifier to enable the field dynamically.",
    )

    field_type = models.TextField(
        choices=FIELD_TYPE_OPTIONS,
        help_text="The form widget to use when displaying the field.",
    )

    field_type_options = JSONField(
        blank=True,
        default=dict,
        help_text="Custom configuration for field types (only used for custom fields).",
        encoder=DjangoJSONEncoder,
    )

    form_field_options = JSONField(
        blank=True,
        default=dict,
        help_text="Custom arguments passed to the form field constructor.",
        encoder=DjangoJSONEncoder,
    )

    form_widget_options = JSONField(
        blank=True,
        default=dict,
        help_text="Custom arguments passed to the form widget constructor.",
        encoder=DjangoJSONEncoder,
    )

    form: "models.ForeignKey[BaseForm, BaseForm]" = FlexibleForeignKey(
        BaseForm, on_delete=models.CASCADE, related_name="fields"
    )
    form_id: Optional[int]

    fieldset_item: "models.OneToOneField[BaseFieldsetItem, BaseFieldsetItem]"
    fieldset_item_id: Optional[int]

    modifiers: "models.BaseManager[BaseFieldModifier]"
    attributes: "models.BaseManager[BaseRecordAttribute]"

    class Meta:
        abstract = True
        order_with_respect_to = "form"
        unique_together = ("form", "name")

    class FlexibleMeta:
        form_field_name = "form"
        form_field_related_name = "fields"

    def __str__(self) -> str:
        return self.label or f"New {cast(str, self._meta.verbose_name).title()}"

    def save(self, *args: Any, **kwargs: Any) -> None:
        """Save the model.

        Sets the machine name if not specified, and ensures that the related
        form is the same as the form that the section belongs to.

        Args:
            args: (Passed to super)
            kwargs: (Passed to super)
        """
        self.name = self.name or slugify(self.label).replace("-", "_")

        super().save(*args, **kwargs)

    def as_field_type(
        self,
        record: Optional["BaseRecord"] = None,
        field_values: Optional[Dict[str, Any]] = None,
    ) -> FieldType:
        """Return the flexible FieldType for the field.

        Args:
            record: The record instance bound to the form.
            field_values: The current values of the fields in the form.

        Returns:
            FieldType: The FieldType class for the field.
        """
        return FIELD_TYPES[self.field_type](
            field=self,
            record=record,
            field_values=field_values,
            modifiers=tuple((m.attribute, m.expression) for m in self.modifiers.all()),
            **self.field_type_options,
        )

    def as_form_field(
        self,
        record: Optional["BaseRecord"] = None,
        field_values: Optional[Dict[str, Any]] = None,
    ) -> forms.Field:
        """Return a Django form Field instance.

        Args:
            field_values: The current values of all fields in the form.
            record: The record instance that the form is bound to, if
                available.

        Returns:
            forms.Field: The configured Django form Field instance.
        """
        return self.as_field_type(
            record=record, field_values=field_values
        ).as_form_field()

    def as_form_widget(
        self,
        record: Optional["BaseRecord"] = None,
        field_values: Optional[Dict[str, Any]] = None,
    ) -> Widget:
        """Return a Django form Widget instance.

        Uses the form_widget_options and the given field_values to customize
        the widget.

        Args:
            record: The record instance bound to the form.
            field_values: The current values of the fields in the form.

        Returns:
            Widget: The configured Django form Widget instance.
        """
        return self.as_field_type(
            record=record, field_values=field_values
        ).as_form_widget()

    def as_model_field(self) -> models.Field:
        """Return a Django model Field definition.

        Returns:
            models.Field: The configured Django model Field instance.
        """
        return self.as_field_type().as_model_field()


class BaseFieldModifier(FlexibleBaseModel):
    """A dynamic expression for customizing field rendering behavior.

    A FieldModifier is essentially a map of `attribute` -> `expression`,
    where `attribute` is the name of a supported attribute on the
    `Field` model, and `expression` is a valid Python expression.

    When `Field.as_form_field()` or `Field.as_model_field()` is called, the
    `expression` is evaluated using the `simpleeval` module with the given
    field_values (usually the field values submitted to the form), and the
    resulting value is used to override the configured attribute on the
    `Field`.

    For example:

    >>> field = Field(required=False)
    >>> unmodified = field.as_form_field()
    >>> repr(unmodified.required)
    'False'
    >>> field.modifiers.create(attribute='required', expression='some_input > 1')
    >>> modified = field.as_form_field({'some_input': 2})
    >>> repr(modified.required)
    'True'

    This is useful for modifying the structure and behavior of forms
    dynamically. Some use cases include:

    * Making fields required only if other fields are filled out.
    * Setting default values based on the values of other fields.
    * Hiding a field until another field changes.
    """

    field: "models.ForeignKey[BaseField, BaseField]" = FlexibleForeignKey(
        BaseField, on_delete=models.CASCADE, related_name="modifiers"
    )
    field_id: Optional[int]

    attribute = models.TextField()
    expression = models.TextField()

    # The _validated flag is used to check whether the expression has been
    # validated. If it's false by the time that save() is called, we validate
    # there.
    _validated = False

    class FlexibleMeta:
        field_field_name = "field"
        field_field_related_name = "modifiers"

    def __str__(self) -> str:
        return f"{cast(str, self._meta.verbose_name).title()} ({self.attribute} = {self.expression})"

    def clean(self) -> None:
        """Ensure that the expression is valid for the form.

        Checks to make sure that referenced names and functions are
        defined before saving.
        """
        super().clean()
        self._validated = False

        # We can only validate the expression if the modifier has been
        # associated with a field and its form (we need to get the initial
        # values from the form to test the expression).
        try:
            field_values = self.field.form.initial_values
        except ObjectDoesNotExist:
            return

        # Validate that the expression is valid for the form. This is
        # accomplished by building a dict of initial values for all fields on
        # the form (along with the initial values for concrete attributes on
        # the Record) and running it through the expression evaluator. If any
        # exceptions are raised, they are returned as validation errors.
        Record = self.flexible_forms.get_model(BaseRecord)
        record_variable = (
            (Record._meta.verbose_name or "record").lower().replace(" ", "_")
        )
        record_attributes = {
            str(f.attname): f.value_from_object(Record())
            for f in cast(List[models.Field], Record._meta.get_fields())
            if hasattr(f, "attname") and hasattr(f, "value_from_object")
        }

        expression_context: Mapping[str, Any] = {
            record_variable: record_attributes,
            **field_values,
        }

        try:
            evaluate_expression(self.expression, names=expression_context)

        # If the expression references a name that isn't a field on the form
        # (or a builtin), raise a validation error.
        except NameNotDefined as ex:
            valid_fields = ", ".join(expression_context.keys())
            name = getattr(ex, "name", "")
            raise ValidationError(
                {
                    "expression": (
                        f"The expression references a variable named '{name}', but no "
                        f"field with that name exists in the form. Valid fields are: "
                        f"{valid_fields}."
                    )
                }
            )

        # If the expression references a function that isn't in scope for the
        # expression, raise a validation error.
        except FunctionNotDefined as ex:
            valid_functions = ", ".join(FormEvaluator.FUNCTIONS.keys())
            func_name = getattr(ex, "func_name", "")
            raise ValidationError(
                {
                    "expression": (
                        f"The expression is trying to use the function {func_name}, "
                        f"but that function does not exist, or cannot be used in "
                        f"expressions. Valid functions are: {valid_functions}"
                    )
                }
            )

        # If the expression encounters another error (e.g., TypeError).
        except BaseException as ex:
            raise ValidationError(
                {
                    "expression": (
                        f"The expression is invalid. Error message: '{str(ex)}'"
                    )
                }
            )

        self._validated = True

    def save(self, *args: Any, **kwargs: Any) -> None:
        """Save the record.

        Runs deferred validation.

        Args:
            args: Passed to super.
            kwargs: Passed to super.
        """
        # If we haven't validated the expression at this point, run clean().
        if not self._validated:
            self.clean()
        super().save(*args, **kwargs)

    class Meta:
        abstract = True
        order_with_respect_to = "field"


class BaseFieldset(FlexibleBaseModel):
    """A section of Fields within a Form."""

    name = models.TextField(
        blank=True,
        default="",
        help_text="The heading to be used for the fieldset. If left empty, no heading will appear.",
    )
    description = models.TextField(
        blank=True, default="", help_text="The description of the fieldset."
    )
    classes = models.TextField(
        blank=True,
        default="",
        help_text="CSS classes to be be applied to the fieldset when rendered.",
    )

    form: "models.ForeignKey[BaseForm, BaseForm]" = FlexibleForeignKey(
        BaseForm, on_delete=models.CASCADE, related_name="fieldsets"
    )
    items: "models.BaseManager[BaseFieldsetItem]"

    class Meta:
        abstract = True

    class FlexibleMeta:
        form_field_name = "form"
        form_field_related_name = "fieldsets"

    def __str__(self) -> str:
        return " ".join(
            filter(
                bool,
                [cast(str, self._meta.verbose_name).title(), self.name, f"({self.pk})"],
            )
        )


class BaseFieldsetItem(FlexibleBaseModel):
    """A single item within a Fieldset."""

    fieldset: "models.ForeignKey[BaseFieldset, BaseFieldset]" = FlexibleForeignKey(
        BaseFieldset, on_delete=models.CASCADE, related_name="items"
    )
    fieldset_id: Optional[int]
    field: "models.OneToOneField[BaseField, BaseField]" = FlexibleOneToOneField(
        BaseField,
        on_delete=models.CASCADE,
        related_name="fieldset_item",
    )
    field_id: Optional[int]

    vertical_order = models.IntegerField(
        help_text=(
            "The vertical order of the item within the fieldset. Items with a lower "
            "vertical order are displayed first. Items with the same vertical order "
            "will be displayed on the same line."
        )
    )
    horizontal_order = models.IntegerField(
        help_text=(
            "The horizontal order of the item within its vertical position. Items "
            "with a lower order are displayed first."
        )
    )

    class Meta:
        abstract = True
        ordering = ["vertical_order", "horizontal_order"]
        unique_together = (
            ("fieldset", "field"),
            ("fieldset", "vertical_order", "horizontal_order"),
        )

    class FlexibleMeta:
        field_field_name = "field"
        field_field_related_name = "fieldset_item"
        fieldset_field_name = "fieldset"
        fieldset_field_related_name = "items"

    def __str__(self) -> str:
        return f"{cast(str, self._meta.verbose_name).title()} {self.pk} for field {self.field_id}"


class RecordManager(models.Manager):
    """A manager for Records.

    Automatically optimizes record retrieval by eagerly loading
    relationships.
    """

    model: Type["BaseRecord"]

    def get_queryset(self) -> "models.QuerySet[BaseRecord]":
        """Define the default QuerySet for fetching Records.

        Eagerly fetches often-used relationships automatically.

        Returns:
            models.QuerySet['BaseRecord']: An optimized queryset of Records.
        """
        RecordAttribute = self.model._flexible_model_for(BaseRecordAttribute)
        return cast(
            "models.QuerySet[BaseRecord]",
            (
                super()
                .get_queryset()
                .select_related("form")
                .prefetch_related(
                    f"form__fields",
                    f"form__fields__modifiers",
                    f"form__fieldsets",
                    f"form__fieldsets__items",
                    Prefetch(
                        "attributes",
                        queryset=RecordAttribute.objects.select_related("field"),
                    ),
                )
            ),
        )


class BaseRecord(FlexibleBaseModel):
    """An instance of a Form.

    The BaseRecord is a bit special. The underlying storage implementation is
    an EAV, but BaseRecord allows callers to access attributes with the usual
    attribute syntax (e.g. record.attribute_name) by overriding __getattr__
    and __setattr__ in order to proxy requests for attributes to the
    underlying BaseRecordAttribute instance(s).

    You might notice that most of the attributes on the BaseRecord are
    prefixed with underscores. This is not necessarily to mark those
    attributes as private, but instead to avoid collisions with EAV attribute
    names.
    """

    form: "models.ForeignKey[BaseForm, BaseForm]" = FlexibleForeignKey(
        BaseForm, on_delete=models.CASCADE, related_name="records"
    )
    form_id: Optional[int]
    attributes: "models.BaseManager[BaseRecordAttribute]"

    objects = RecordManager()

    # The _initialized flag is used to determine if the model instance has been
    # fully initialized by Django. We use the flag to determine if it's safe to
    # start proxying __getattr__ and __setattr__ calls to our attributes
    # association (which breaks if the instance is not fully initialized).
    _initialized: bool = False

    # The __setattr__ proxy stores attribute updates in the _unsaved_changes
    # dict until save() is called. This attempts to mirror the way vanilla
    # Django models work.
    _unsaved_changes: Dict[str, Any]

    class Meta:
        abstract = True

    class FlexibleMeta:
        form_field_name = "form"
        form_field_related_name = "records"

    def __str__(self) -> str:
        # If we can get the name of the record's form without making a database
        # call, do it so we can make the __str__ more descriptive. Otherise,
        # just call it a "Record".
        record_type = (
            self.form.label
            if self.__dict__.get("form")
            or self._meta.get_field("form").get_cached_value(self, default=None)
            else cast(str, self._meta.verbose_name).title()
        )

        return f"New {record_type}" if not self.pk else f"{record_type} {self.pk}"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._unsaved_changes = {}
        self._initialized = True

    def as_django_fieldsets(self) -> Sequence[DjangoFieldset]:
        """Generate Django fieldsets from the BaseRecord instance.

        Returns:
            Sequence[DjangoFieldset]: Django-compatible fieldsets for the instance.
        """
        pre_fieldsets_prepare.send(sender=self.__class__, instance=self)
        django_fieldsets = self.form.as_django_fieldsets()
        post_fieldsets_prepare.send(
            sender=self.__class__, instance=self, django_fieldsets=django_fieldsets
        )

        return django_fieldsets

    def as_django_form(self, *args: Any, **kwargs: Any) -> "BaseRecordForm":
        """Create a Django form from the record.

        Uses `self` as the `instance` argument to as_django_form()`

        Args:
            args: Passed to super.
            kwargs: Passed to super.

        Returns:
            BaseRecordForm: A configured Django form for the record.
        """
        kwargs = {"instance": self, **kwargs}
        return self.form.as_django_form(*args, **kwargs)

    @cached_property
    def _fields(self) -> Mapping[str, BaseField]:
        """Return a map of Fields for the Record's form, keyed by their names.

        Returns:
            Mapping[str, BaseField]: A mapping of Field instances by their
                names.
        """
        if not hasattr(self, "form"):
            return {}
        return {f.name: f for f in self.form.fields.all()}

    @property
    def _data(self) -> Dict[str, Any]:
        """Return a dict of Record attributes and their values.

        Returns:
            Mapping[str, Any]: A dict of Record attributes and their values.
        """
        initial_values = {name: field.initial for name, field in self._fields.items()}
        attribute_values = cast(
            Dict[str, Any],
            ({a.field.name: a.value for a in self.attributes.all()} if self.pk else {}),
        )
        staged_attributes = {k: v for k, v in self._unsaved_changes.items()}

        return {
            **initial_values,
            **attribute_values,
            **staged_attributes,
        }

    def __getattr__(self, name: str) -> Any:
        """Get an attribute value from the record.

        Overrides __getattr__ to make attribute access more natural-feeling.
        Attributes on the record can be accessed with obj.attribute syntax.

        If the record hasn't finished initialization yet, falls back to
        default behavior.

        Args:
            name: The name of the attribute to get from the record.

        Raises:
            AttributeError: If the object has no attribute with the given
                name.

        Returns:
            Any: The value of the attribute, if available.
        """
        if (
            not self._initialized
            or name
            in frozenset(
                [
                    *self.__dict__.keys(),
                    *self.__class__.__dict__.keys(),
                ]
            )
            or name.startswith("_")
        ):
            return self.__getattribute__(name)

        if name in self._fields:
            return self._data[name]

        # HACK: Choice fields usually add a get_FIELDNAME_display method, which
        # we handle here by rendering the label from the value in _data.
        display_method = re.match(r"^get_(?P<field_name>\w+)_display$", name)
        if display_method:

            def _get_choice_display(
                self_: BaseRecord = self,
                display_method: Match[str] = cast(Match[str], display_method),
            ) -> Optional[str]:
                field_name = display_method.group("field_name")

                try:
                    form_field = self_._fields[field_name].as_form_field(
                        record=self_, field_values=self_._data
                    )
                    return next(
                        (
                            cast(str, label)
                            for value, label in form_field.choices
                            if value == self_._data.get(field_name)
                        ),
                        None,
                    )
                except (KeyError, AttributeError) as ex:
                    raise AttributeError(
                        f"'{self_.__class__.__name__}' object has no attribute '{name}'"
                    ) from ex

            return _get_choice_display

        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )

    def __setattr__(self, name: str, value: Any) -> None:
        """Set an attribute value by name.

        Args:
            name: The name of the attribute to set.
            value: The value that the attribute should be set to.
        """
        if (
            not self._initialized
            or name
            in frozenset([*self.__dict__.keys(), *self.__class__.__dict__.keys(), "pk"])
            or name.startswith("_")
        ):
            super().__setattr__(name, value)
            return

        RecordAttribute = cast(Any, self._flexible_model_for(BaseRecordAttribute))
        self._unsaved_changes[name] = RecordAttribute(
            record=self,
            field=self._fields[name],
            value=value,
        ).value
        self._invalidate_caches("_data")

    def _invalidate_caches(self, *caches: str) -> None:
        """Invalidate cached properties on the Record."""
        caches = caches or ("_data", "_fields")

        for cache in caches:
            try:
                delattr(self, cache)
            except AttributeError:
                pass

    @transaction.atomic
    def save(self, *args: Any, **kwargs: Any) -> None:
        """Save the record and invalidate the property caches."""
        super().save(*args, **kwargs)

        # If there are no attributes to update, return early.
        if not self._unsaved_changes:
            return

        # Upsert the record attributes.
        RecordAttribute = cast(Any, self._flexible_model_for(BaseRecordAttribute))
        attribute_map = {a.field.name: a for a in self.attributes.all()}

        value_fields: Set[str] = set()
        update: List[BaseRecordAttribute] = []
        insert: List[BaseRecordAttribute] = []
        for field_name, value in self._unsaved_changes.items():
            # Find the existing attribute object or create a new one.
            attribute = attribute_map.get(field_name) or RecordAttribute(
                record=self,
                field=self._fields[field_name],
            )

            # Set the value for the attribute.
            attribute.value = value

            # Add the attribute object to the appropriate operation set.
            operation = update if attribute.pk else insert
            operation.append(attribute)

            # If we're updating, track which value column should be updated.
            if operation is update:
                value_fields.add(attribute.value_field_name)

        # Perform bulk updates and inserts as necessary.
        if update:
            RecordAttribute._default_manager.bulk_update(
                update, fields=(*value_fields,)
            )
        if insert:
            RecordAttribute._default_manager.bulk_create(insert)

        # Invalidate the data cache.
        self._invalidate_caches("_data")


class BaseRecordAttribute(FlexibleBaseModel):
    """A value for an attribute on a single Record."""

    ##
    # _VALUE_FIELD_PREFIX
    #
    # A prefix used to denote value fields for the attribute.
    #
    _VALUE_FIELD_PREFIX = "_value_"

    record: "models.ForeignKey[BaseRecord, BaseRecord]" = FlexibleForeignKey(
        BaseRecord,
        on_delete=models.CASCADE,
        related_name="attributes",
    )
    record_id: Optional[int]

    field: "models.ForeignKey[BaseField, BaseField]" = FlexibleForeignKey(
        BaseField,
        on_delete=models.CASCADE,
        related_name="attributes",
    )
    field_id: Optional[int]

    class Meta:
        abstract = True

    class FlexibleMeta:
        field_field_name = "field"
        field_field_related_name = "attributes"
        record_field_name = "record"
        record_field_related_name = "attributes"

    def __str__(self) -> str:
        return f"{cast(str, self._meta.verbose_name).title()} {self.pk}"

    @classmethod
    def get_value_field_name(cls, field_type: str) -> str:
        """Return the name of the field used for storing field_type values.

        Args:
            field_type: The field type.

        Returns:
            str: The name of the model field for storing field_type values.
        """
        return f"{cls._VALUE_FIELD_PREFIX}{field_type}"

    @property
    def value_field_name(self) -> str:
        """Return the name of the field that holds the attribute value.

        Values are stored in columns with appropriate typings for performance
        and lossless storage.

        Returns:
            str: The name of the field that holds the value for the attribute.
        """
        return self.get_value_field_name(self.field.field_type)

    def _get_value(self) -> Any:
        """Return the attribute value.

        Returns:
            Any: The current value of the attribute.
        """
        return getattr(self, self.value_field_name)

    def _set_value(self, new_value: Any) -> None:
        """Set the value of the attribute.

        Sets the value of the attribute to the given value.

        Args:
            new_value: The new value for the attribute.
        """
        # Clear out the
        value_field_names = (
            f.name
            for f in self._meta.get_fields()
            if f.name.startswith(self._VALUE_FIELD_PREFIX)
        )
        for field_name in value_field_names:
            setattr(self, field_name, None)

        setattr(self, self.value_field_name, new_value)

    ##
    # value
    #
    # A getter and setter property for transparently interacting with the value
    # of the attribute.
    #
    value = property(_get_value, _set_value)


@receiver(class_prepared)
def update_record_attribute_model(sender: Type[models.Model], **kwargs: Any) -> None:
    """Add model fields for every FieldType to BaseRecordAttribute.

    The EAV pattern used for storing form submissions (Records) achieves lossless
    storage by creating a column of an appropriate datatype for each supported
    field.

    Args:
        sender: The model class that was prepared.
        kwargs: Unused, but required.
    """
    # If the sender isn't a BaseRecordAttribute, or the sender is an abstract
    # model, noop; we only want to operate on concrete models that inherit from
    # BaseRecordAttribute.
    if not issubclass(sender, BaseRecordAttribute) or sender._meta.abstract:
        return

    for field_type_name, field_type in sorted(FIELD_TYPES.items(), key=lambda f: f[0]):
        sender.add_to_class(
            sender.get_value_field_name(field_type_name),
            field_type.as_model_field(blank=True, null=True, default=None),
        )


class FlexibleForms:
    """A class for generating new sets of concrete flexible form models.

    Enables users to have multiple sets of FlexibleForm models in their
    application.

    Effectively frees the developer from having to shoehorn all of their use
    cases into a single set of models and tables, and allows them to
    customize each set of models and tailor them to each specific use case.

    For example, in a school setting it might be useful to have a flexible
    model structure for Quizzes, and another one for Projects, each with
    their own set of models, database tables, and attributes. The developer
    can also choose to rename the relationships that connect these models for
    ergonomics:

    ```python
    # quizzes/models.py
    from django.contrib.auth import get_user_model
    from django.db import models
    from flexible_forms.models import BaseForm, BaseRecord
    from django.utils import timezone

    quizzes = FlexibleForms(model_prefix="Quiz")

    class QuizAssignment(quizzes.BaseForm):
        due = models.DateTimeField(default=timezone.now)

        class Meta(BaseForm.Meta):
            ordering = ("-due",)

    class QuizSubmission(quizzes.BaseRecord):
        student = models.ForeignKey(get_user_model(), on_delete=models.CASCADE)
        submitted = models.DateTimeField(default=timezone.now)
        grade = models.PositiveIntegerField()

        class FlexibleMeta:
            form_field_name = "quiz"
            form_field_related_name = "submissions"
    ```

    ```python
    # projects/models.py
    from django.db import models
    from flexible_forms.models import BaseForm
    from django.utils import timezone

    projects = FlexibleForms(model_prefix="Project")

    class ProjectAssignment(projects.BaseForm):
        due = models.DateTimeField(default=timezone.now)
        grading_criteria = models.TextField()

        class Meta(BaseForm.Meta):
            ordering = ("-due",)
    ```
    """

    finalized: bool = False

    def __init__(self, model_prefix: Optional[str] = None) -> None:
        """Initialize the FlexibleForms object.

        Args:
            model_prefix: The prefix to use when generating model names for
                default models.
        """
        self.model_prefix = model_prefix or ""
        self.module = cast(ModuleType, inspect.getmodule(inspect.stack()[1][0]))
        self.models: Dict[
            Union[Type[FlexibleBaseModel], str], Type[FlexibleBaseModel]
        ] = {}
        self._finalizer = weakref.finalize(self, self._check_finalized, self)

    def make_flexible(self) -> None:
        """Generate default models.

        FlexibleForms will generate default models for any models that
        were not explicity implemented in the source module.
        """
        for base_model in FlexibleBaseModel.__subclasses__():
            self.get_model(base_model, create=True)
        self.finalized = True

    def register_model(self, model_class: Type[FlexibleBaseModel]) -> None:
        """Register a concrete implementation of a flexible base model.

        Maps the given model_class' base model class (e.g.,
        flexible_forms.models.BaseForm) the concrete implementation
        (model_class).

        Additionally, adds a string entry (e.g. "flexible_forms.baseform")
        that points to the same concrete implementation.

        Args:
            model_class: The concrete implementation of a flexible base model.
        """
        self.models[self._get_flexible_base_model(model_class)] = model_class

    def get_model(
        self,
        base_model: Type[M],
        *,
        create: bool = False,
    ) -> Type[M]:
        """Returns the concrete model for the given flexible base model.

        Args:
            base_model: The base model for which to return the concrete model
                implementation.
            create: When True, generates a default model with the given base
                class if one does not already exist.

        Raises:
            LookupError: if no concrete model has been registered for the
                given base yet.

        Returns:
            Type[FlexibleBaseModel]: The concrete model implementation for
                the base model.
        """
        # If no model is registered for the requested base_model yet, and
        # create is True, create and register a new model.
        if base_model not in self.models and create:
            self.register_model(
                type(
                    base_model.__name__.replace("Base", self.model_prefix),
                    (getattr(self, base_model.__name__),),
                    {"__module__": self.module.__name__},
                )
            )

        try:
            return cast(Type[M], self.models[base_model])
        except KeyError:
            raise LookupError(
                f"No concrete model has been registered for {base_model}: {self.models}"
            )

    def _get_flexible_base_model(
        self, model_class: Type["FlexibleBaseModel"]
    ) -> Type["FlexibleBaseModel"]:
        """Resolve the flexible base model for the given model_class.

        Searches the __bases__ of the given model class for one that
        corresponds to one of the core flexible base models.

        Args:
            model_class: The model class for which to resolve the flexible
                base class.

        Returns:
            Type[FlexibleBaseModel]: The flexible base model.

        Raises:
            ValueError: if the model_class does not inherit from a flexible
                base class.
        """
        flexible_bases = frozenset(FlexibleBaseModel.__subclasses__())
        for base in model_class.__bases__:
            try:
                return next(
                    flexible_base
                    for flexible_base in flexible_bases
                    if issubclass(base, flexible_base)
                )
            except StopIteration:
                continue

        raise ValueError(
            f"{model_class.__name__} does not inherit from FlexibleBaseModel."
        )

    @property
    def BaseForm(self) -> Type["BaseForm"]:
        """Return a BaseForm with a reference to flexible_forms.

        Returns:
            Type[BaseForm]: A BaseForm class with a flexible_forms
                attribute.
        """
        return self._generate_model(BaseForm)

    @property
    def BaseField(self) -> Type["BaseField"]:
        """Return a BaseField with a reference to flexible_forms.

        Returns:
            Type[BaseField]: A BaseField class with a flexible_forms
                attribute.
        """
        return self._generate_model(BaseField)

    @property
    def BaseFieldset(self) -> Type["BaseFieldset"]:
        """Return a BaseFieldset with a reference to flexible_forms.

        Returns:
            Type[BaseFieldset]: A BaseFieldset class with a flexible_forms
                attribute.
        """
        return self._generate_model(BaseFieldset)

    @property
    def BaseFieldsetItem(self) -> Type["BaseFieldsetItem"]:
        """Return a BaseFieldsetItem with a reference to flexible_forms.

        Returns:
            Type[BaseFieldsetItem]: A BaseFieldsetItem class with a
                flexible_forms attribute.
        """
        return self._generate_model(BaseFieldsetItem)

    @property
    def BaseFieldModifier(self) -> Type["BaseFieldModifier"]:
        """Return a BaseFieldModifier with a reference to flexible_forms.

        Returns:
            Type[BaseFieldModifier]: A BaseFieldModifier class with a
                flexible_forms attribute.
        """
        return self._generate_model(BaseFieldModifier)

    @property
    def BaseRecord(self) -> Type["BaseRecord"]:
        """Return a BaseRecord with a reference to flexible_forms.

        Returns:
            Type[BaseRecord]: A BaseRecord class with a flexible_forms
                attribute.
        """
        return self._generate_model(BaseRecord)

    @property
    def BaseRecordAttribute(self) -> Type["BaseRecordAttribute"]:
        """Return a BaseRecordAttribute with a reference to flexible_forms.

        Returns:
            Type[BaseRecordAttribute]: A BaseRecordAttribute class with a
                flexible_forms attribute.
        """
        return self._generate_model(BaseRecordAttribute)

    def _generate_model(self, base_model: Type[M]) -> Type[M]:
        """Generate a model class from the given base_model.

        Args:
            base_model: The base model for the new model.

        Returns:
            Type[T]: A new model with a reference to flexible_forms.
        """
        return type(
            f"{self.model_prefix}{base_model.__name__}",
            (base_model,),
            {
                "__module__": self.module.__name__,
                "flexible_forms": self,
                "Meta": type(
                    "Meta",
                    (base_model.Meta,),
                    {
                        "abstract": True,
                    },
                ),
            },
        )

    @staticmethod
    def _check_finalized(flexible_forms: "FlexibleForms") -> None:
        """Ensure that make_flexible has been called.

        Attempts to catch a developer mistake where they forget to call
        make_flexible after defining their model overrides.

        Args:
            flexible_forms: The FlexibleForms object to be checked.

        Raises:
            ImproperlyConfigured: If make_flexible has not been called on the
                object before exiting.
        """
        if flexible_forms.finalized:
            return
        raise ImproperlyConfigured(
            f"A FlexibleForms object was created in {flexible_forms.module}, but "
            f"`make_flexible` was never called on it."
        )



# -*- coding: utf-8 -*-
from django.urls import path

from . import views

app_name = "flexible_forms"
urlpatterns = [
    path(
        "<str:app_label>/<str:model_name>/<int:field_pk>/autocomplete",
        views.autocomplete,
        name="autocomplete",
    ),
]



# -*- coding: utf-8 -*-
from django.conf import settings
from django.core.cache import DEFAULT_CACHE_ALIAS, caches

cache = caches[getattr(settings, "FLEXIBLE_FORMS_CACHE", DEFAULT_CACHE_ALIAS)]
cache_prefix = "flexible_forms"



# -*- coding: utf-8 -*-
import json
from decimal import Decimal
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from django.contrib.admin.widgets import AutocompleteMixin
from django.forms.widgets import Select

from flexible_forms.utils import stable_json


class AutocompleteSelect(Select):
    """A select widget where options are sourced from a remote URL."""

    url: Optional[str]
    allow_freetext: bool
    placeholder: str

    EMPTY_VALUES = (None, "", "null")

    def __init__(
        self,
        url: Optional[str] = None,
        cacheable: bool = True,
        allow_freetext: bool = False,
        placeholder: Optional[str] = None,
        attrs: Optional[Dict[str, Any]] = None,
        choices: Sequence[Tuple[Any, Any]] = (),
        **kwargs: Any,
    ) -> None:
        super().__init__(attrs=attrs, choices=choices)
        self.url = url
        self.cacheable = cacheable
        self.allow_freetext = allow_freetext
        self.placeholder = placeholder or ""

    def optgroups(  # type: ignore
        self, name: str, value: str, attrs: Optional[Dict[Any, Any]] = None
    ) -> List[Tuple[None, List[Dict[str, Any]], int]]:
        """Build a list of optgroups for populating the select.

        Autocompletes are powered by external endpoints, and so the selected
        options must be derived from the selected value to avoid making an
        HTTP call just to render the field.

        The AutocompleteWidget assumes that the stored value is the JSON
        representation of the originally-selected result from the
        autocomplete API call.

        Args:
            name: The name of the widget in the form.
            value: The value of the field.
            attrs: Unused.

        Returns:
            List[Tuple[Optional[str], List[Tuple[Any, Any]], int]]: Optgroups
                for the widget.
        """
        default: Tuple[None, List[Dict[str, Any]], int] = (None, [], 0)
        groups = [default]
        has_selected = False

        # Parse the given value as a JSON string, and make it a list if it
        # isn't already so that we can iterate over it when we enumerate
        # selected options.
        parsed_value = json.loads(value)
        parsed_value = (
            parsed_value if isinstance(parsed_value, list) else [parsed_value]
        )

        selected_choices = {
            stable_json(v)
            for v in parsed_value
            if not self._choice_has_empty_value((stable_json(v), ""))
        }

        if not self.is_required and not self.allow_multiple_selected:
            default[1].append(self.create_option(name, "", "", False, 0))

        choices = [
            (stable_json(v), v["text"])
            for v in parsed_value
            if v not in (None, "", "null")
        ]

        for option_value, option_label in choices:
            selected = str(option_value) in parsed_value and (
                has_selected is False or self.allow_multiple_selected
            )
            has_selected |= selected
            index = len(default[1])
            subgroup = default[1]
            subgroup.append(
                self.create_option(
                    name, option_value, option_label, selected_choices, index
                )
            )

        return groups

    def build_attrs(
        self, base_attrs: Dict[str, Any], extra_attrs: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Union[Decimal, float, str]]:
        """Set select2's AJAX attributes.

        Attributes can be set using the html5 data attribute.
        Nested attributes require a double dash as per
        https://select2.org/configuration/data-attributes#nested-subkey-options
        """
        attrs = super().build_attrs(base_attrs, extra_attrs=extra_attrs)
        return {
            **attrs,
            "data-ajax--cache": json.dumps(self.cacheable),
            "data-ajax--delay": 250,
            "data-ajax--type": "GET",
            "data-ajax--url": self.url or "",
            "data-theme": "admin-autocomplete",
            "data-allow-clear": json.dumps(not self.is_required),
            "data-placeholder": self.placeholder,
            "data-close-on-select": "false" if self.allow_multiple_selected else "true",
            "data-tags": json.dumps(self.allow_freetext),
            "data-disabled": json.dumps(
                "true"
                in (
                    str(attrs.get("disabled", "false")).lower(),
                    str(attrs.get("readonly", "false")).lower(),
                )
            ),
            "class": " ".join(
                [str(attrs.get("class", "")), "admin-autocomplete"]
            ).strip(),
        }

    def value_from_datadict(
        self, data: Dict[str, Any], files: Mapping[str, Iterable[Any]], name: str
    ) -> str:
        """Extract the widget's value from the given data.

        Handles the case where multiple values are selected.

        Args:
            data: The form data.
            files: File uploads submitted with the form data.
            name: The name of the widget.

        Returns:
            str: The JSON value produced by the widget.
        """
        value = super().value_from_datadict(data, files, name)

        if value is None:
            value = []

        if not isinstance(value, list):
            value = [value]

        parsed_value: List[dict] = []
        for v in value:
            # Skip empty values.
            if v in self.EMPTY_VALUES:
                continue

            parsed_v = None

            # If the value is already a dict, we'll just consider it to be
            # already parsed.
            if isinstance(v, dict):
                parsed_v = v

            # If the value is a string and it looks like a JSON object, parse
            # it as one.
            elif isinstance(v, str):
                if v.startswith("{") and v.endswith("}"):
                    parsed_v = json.loads(v)
                else:
                    freetext_option = {"text": v, "value": v, "extra": {}}
                    parsed_v = {**freetext_option, "id": stable_json(freetext_option)}

            if parsed_v is not None:
                parsed_value.append(parsed_v)

        # If only one value is expected, unwrap it from the parsed list.
        final_value = (
            next(iter(parsed_value), None)
            if not self.allow_multiple_selected
            else parsed_value
        )

        return stable_json(final_value)

    def format_value(self, value: Any) -> Any:
        """Format a value for rendering in a widget template.

        Args:
            value: The value to format.

        Returns:
            Any: The formatted value, if formatting was successful. Otherwise
                returns the value unmodified.
        """
        return str([] if value in self.EMPTY_VALUES else value)

    @staticmethod
    def _choice_has_empty_value(choice: Tuple[Any, str]) -> bool:
        """Return True if the value of the given choice is empty.

        Args:
            choice: The choice two-tuple to check for an empty value.

        Returns:
            bool: True if the value of the given choice is empty, Flase
                otherwise.
        """
        value, _ = choice
        return value in AutocompleteSelect.EMPTY_VALUES

    # Use the media property from Django Admin's AutocompleteMixin (which this
    # widget is based on). This ensures that the Django admin appropriately
    # loads the required Select2 frontend assets.
    media = AutocompleteMixin.media


class AutocompleteSelectMultiple(AutocompleteSelect):
    """An AutocompleteSelect that supports multiple selections."""

    allow_multiple_selected = True



# -*- coding: utf-8 -*-
"""Forms that power the flexible_forms module."""

import datetime
from typing import Any, Dict, Mapping, Optional, cast

import django
from django import forms
from django.core.exceptions import ValidationError
from django.core.files.base import File
from django.forms.models import ALL_FIELDS
from django.forms.widgets import HiddenInput
from django.utils.datastructures import MultiValueDict

from flexible_forms.cache import cache
from flexible_forms.models import BaseForm, BaseRecord
from flexible_forms.signals import (
    post_form_clean,
    post_form_init,
    post_form_save,
    pre_form_clean,
    pre_form_init,
    pre_form_save,
)

if django.VERSION >= (3, 1):  # pragma: no-cover
    from django.forms.models import ModelChoiceIteratorValue  # type: ignore


class BaseRecordForm(forms.ModelForm):
    """A Django form for serializing Form submissions into Record objects.

    It is used as the base class when generating a Django Form from a
    Form object.
    """

    class Meta:
        fields = ALL_FIELDS
        model: "BaseRecord"

    def __init__(
        self,
        data: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, File]] = None,
        instance: Optional["BaseRecord"] = None,
        initial: Optional[Mapping[str, Any]] = None,
        prefix: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        opts = self._meta  # type: ignore

        # Set the prefix if we were given one (we'll need it for a modifier).
        self.prefix = prefix

        # Ensure that data and files are both mutable so that signal handlers
        # can act before the form is initialized.
        data = data.copy() if data is not None else None
        files = files.copy() if files is not None else None

        # If we don't have an instance (e.g., we're adding a new record), we try
        # to derive the BaseForm from the given parameters.
        #
        # If we can't derive a BaseForm (e.g. we're totally unbound and there's
        # no BaseForm identifier in the initial data) then we behave as a
        # totally normal ModelForm, only presenting the BaseForm's concrete
        # attributes.
        RecordModel = opts.model
        FormModel = RecordModel._flexible_model_for(BaseForm)
        form_field_name = RecordModel.FlexibleMeta.form_field_name

        # Try to get some kind of specifier for the BaseForm we should use.
        #
        #   1. Look in the data parameter to see if one was sumbitted with the
        #      form data.
        #   2. If it wasn't in the form data, see if the instance is related to
        #      a BaseForm.
        #   3. Look in the initial parameter to see if we were given a BaseForm
        #      either manually or by the as_django_form() method on the BaseForm
        #      model.
        #
        # If all of these fail, we'll fall back to behaving as a normal
        # ModelForm: the form will only have fields for the direct model
        # attributes of a BaseRecord until it has a relationship to a BaseForm.
        #
        form = (data or {}).get(form_field_name)
        form = form or getattr(instance, form_field_name, None)
        form = form or (initial or {}).get(form_field_name)

        # Depending on how the RecordForm is being created, the form might be:
        #
        #   3. A ModelChoiceIteratorValue (Django 3.1+), which has an "instance"
        #      property containing the form instance that we unpack.
        #   4. A primary key value (an int or a string) that we use to query
        #      for the BaseForm.
        #   1. None or a BaseForm instance, in which case we do nothing.
        if django.VERSION >= (3, 1) and isinstance(form, ModelChoiceIteratorValue):
            form = form.instance
        if isinstance(form, int) or (isinstance(form, str) and form.isdigit()):
            form = FormModel.objects.get(pk=form)

        # If the form is bound, make sure that data holds a reference to the
        # form object, and disable the form field.
        is_bound = data is not None or files is not None
        if is_bound:
            data = cast(Dict[str, Any], data or MultiValueDict())
            data[form_field_name] = form

        # Inject the instance's _data (form field values) into the initial dict.
        # If we weren't given an instance, we make a new one (but don't persist
        # it) for consistency.
        instance = instance or opts.model(**{form_field_name: form})
        initial = {
            **instance._data,
            **(initial or {}),
            form_field_name: form,
        }

        # If any of the form fields have a "_value" attribute, use it in either
        # the data (if the form is bound) and/or the initial (if the form is
        # unbound).
        modified_fields = {
            k: v for k, v in self.base_fields.items() if hasattr(v, "_value")
        }
        for field_name, field in modified_fields.items():
            field_name = self.add_prefix(field_name)

            try:
                field_value = field._value  # type: ignore
            except AttributeError:
                continue

            # Set the initial value.
            initial[field_name] = field_value

            # For unbound forms, data and files are both None, so we can't set
            # values in them and we continue on.
            if not is_bound:
                continue

            files = cast(Dict[str, File], files or MultiValueDict())
            data = cast(Dict[str, Any], data or MultiValueDict())
            value = data.get(field_name, files.get(field_name))

            # If the field was already assigned a non-empty value, don't try to
            # overwrite it.
            if value not in field.empty_values:
                continue

            # Set the appropriate data element (files for FileFields, data for
            # everything else) to the field's new value.
            if isinstance(field, forms.FileField):
                files[field_name] = field_value
            else:
                data[field_name] = field_value

            # Unset the initial value so that the automatically-set value is
            # detected as a change when the form is saved.
            initial.pop(field_name, None)

        # Emit a signal before initializing the form.
        pre_form_init.send(
            sender=self.__class__,
            form=self,
            data=data,
            files=files,
            instance=instance,
            initial=initial,
        )

        # Initialize the form as usual.
        super().__init__(
            data=data, files=files, instance=instance, initial=initial, **kwargs
        )

        # Hide and disable the form input if the BaseRecord is already persisted
        # with a relationship to its BaseForm.
        if form is not None and form_field_name in self.fields:
            form_field = self.fields[form_field_name]
            form_field.widget = HiddenInput()
            form_field.disabled = instance.pk and getattr(
                instance, f"{form_field_name}_id", None
            )

        # Emit a signal after initializing the form.
        post_form_init.send(
            sender=self.__class__,
            form=self,
        )

    def full_clean(self) -> None:
        """Perform a full clean of the form.

        Emits signals before and after, and excludes the form relation
        from validation if it's already set (to eliminate database
        queries related to the schema lookup).
        """
        pre_form_clean.send(sender=self.__class__, form=self)

        super().full_clean()

        record_pk = self.instance.pk
        record_opts = self.instance._meta
        app_label, model_name = record_opts.app_label, record_opts.model_name

        field_values = {
            **{
                name: field.widget.value_from_datadict(
                    self.data, self.files, self.add_prefix(name)
                )
                for name, field in self.fields.items()
            },
            **getattr(self, "cleaned_data", {}),
        }

        for key, value in field_values.items():
            if isinstance(value, File):
                field_values[key] = value.name

        cache.set(
            f"flexible_forms:field_values:{app_label}:{model_name}:{record_pk}",
            field_values,
            timeout=None,
        )

        clean_responses = post_form_clean.send_robust(
            sender=self.__class__, form=self, field_values=field_values
        )

        for _, response in clean_responses:
            if not isinstance(response, BaseException):
                continue
            if isinstance(response, ValidationError):
                self.add_error(None, response)
            raise response

    def clean(self) -> Dict[str, Any]:
        """Clean the form data before saving."""
        # Emit a signal before initializing the form.

        cleaned_data = super().clean()

        for key, value in cleaned_data.items():
            field = self.fields.get(key)

            # Time-only values cannot be timezone aware, so we remove the
            # timezone if one is given.
            if isinstance(value, datetime.time):
                value = value.replace(tzinfo=None)

            # If a False value is given for a FileField, it means that the file
            # should be cleared, and its value set to None.
            if isinstance(field, forms.FileField) and value is False:
                value = None

            # Replace the value in cleaned_data with the updated value.
            cleaned_data[key] = value

        return cleaned_data

    def save(self, commit: bool = True) -> "BaseRecord":
        """Save the form data to a Record.

        Maps the cleaned form data into the Record's _data field.

        Args:
            commit: If True, persists the data to the database.

        Returns:
            instance: The Record model instance.
        """
        pre_form_save.send(sender=self.__class__, form=self)

        # Update any changed attributes.
        for field_name in self.changed_data:
            if field_name not in self.cleaned_data:
                continue
            setattr(self.instance, field_name, self.cleaned_data[field_name])

        super().save(commit=commit)

        post_form_save.send(sender=self.__class__, form=self)

        return cast("BaseRecord", self.instance)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flexible_forms']

package_data = \
{'': ['*'], 'flexible_forms': ['static/flexible_forms/admin/*']}

install_requires = \
['django>=2.2',
 'jmespath>=0.10.0,<0.11.0',
 'requests>=2.24.0,<3.0.0',
 'simpleeval>=0.9.10,<0.10.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata']}

setup_kwargs = {
    'name': 'django-flexible-forms',
    'version': '0.16.12',
    'description': 'A reusable Django app for managing database-backed forms.',
    'long_description': None,
    'author': 'Eric Abruzzese',
    'author_email': 'eric.abruzzese@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-

"""Flexible forms.

A Django app to support configuration-driven forms backed by the
database.
"""

from django.apps import AppConfig


class FlexibleFormsConfig(AppConfig):
    """A Django admin configuration for flexible_forms.

    Handles initialization logic for the flexible_forms app.
    """

    name = "flexible_forms"
    verbose_name = "Flexible Forms"



# -*- coding: utf-8 -*-

"""Django admin configurations for flexible_forms."""

import json
import logging
from typing import Any, Iterable, Mapping, Optional, Sequence, Type, cast

from django import forms
from django.conf import settings
from django.contrib import admin
from django.contrib.admin.options import InlineModelAdmin
from django.db import models
from django.http import HttpRequest
from django.urls import reverse
from django.utils.safestring import SafeText, mark_safe

from flexible_forms.models import (
    AliasField,
    BaseField,
    BaseFieldModifier,
    BaseFieldset,
    BaseFieldsetItem,
    BaseForm,
    BaseRecord,
    DjangoFieldset,
    FlexibleBaseModel,
    JSONField,
)

logger = logging.getLogger(__name__)

##
# Nested admin support
#
# If the project has the nested_admin Django app installed, we'll try to use it
# so that field modifiers can be configured via the Django Admin.
#
if "nested_admin" in settings.INSTALLED_APPS:
    from nested_admin.nested import NestedModelAdmin as ModelAdmin
    from nested_admin.nested import NestedStackedInline as StackedInline
    from nested_admin.nested import NestedTabularInline as TabularInline
else:  # pragma: no cover
    ModelAdmin = admin.ModelAdmin
    StackedInline = admin.StackedInline
    TabularInline = admin.TabularInline


DEFAULT_FORMFIELD_OVERRIDES = cast(
    Mapping[Type[models.Field], Mapping[str, Any]],
    {
        models.TextField: {
            "widget": forms.widgets.TextInput(
                attrs={
                    "size": "50",
                },
            ),
        },
    },
)


##
# Ace editor support
#
# If the project has django_ace installed, we'll try to use it to render JSON
# fields using a friendlier editor.
#
if "django_ace" in settings.INSTALLED_APPS:
    from django_ace import AceWidget

    class CodeEditorAdminWidget(AceWidget):
        """An admin widget for editing code.

        Powered by Ace editor.
        """

        class Media:
            css = {"all": ("flexible_forms/admin/code-editor-admin-widget.css",)}

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            super().__init__(
                *args,
                **{
                    "height": "auto",
                    "maxlines": "Infinity",
                    "toolbar": False,
                    **kwargs,
                },
            )

        def format_value(self, value: Any) -> Optional[str]:
            """Format the given value for use by the widget.

            Args:
                value: The value to format.

            Returns:
                str: The formatted value.
            """
            try:
                return json.dumps(json.loads(value), indent=2, sort_keys=True)
            except BaseException:
                return cast(Optional[str], super().format_value(value))

    # Use the code editor for JSONField fields.
    DEFAULT_FORMFIELD_OVERRIDES = {
        **DEFAULT_FORMFIELD_OVERRIDES,
        JSONField: {"widget": CodeEditorAdminWidget(mode="json")},
    }


class FlexibleAdminMixin:
    """An admin class mixin for flexible form models."""

    model: Type[FlexibleBaseModel]
    formfield_overrides = DEFAULT_FORMFIELD_OVERRIDES

    @property
    def exclude(self) -> Sequence[str]:
        """Exclude alias fields from the admin form.

        Handles the case where the implementer has aliased one or more fields
        on the model.

        Returns:
            Tuple[str]: A tuple with the names of aliased fields.
        """
        return tuple(
            f.name for f in self.model._meta.fields if isinstance(f, AliasField)
        )


class FieldModifiersInline(FlexibleAdminMixin, TabularInline):
    """An inline representing a single modifier for a field on a form."""

    classes = ("collapse",)
    model = BaseFieldModifier
    extra = 1

    @property
    def fk_name(self) -> str:
        """Return the name of the field with a ForeignKey to the Field.

        Handles the case where the implementer has aliased the "field" field.

        Returns:
            str: The name of the field ForeignKey field.
        """
        original_fk_name = BaseFieldModifier.FlexibleMeta.field_field_name
        concrete_fk_name = next(  # pragma: no cover
            (
                f.original_field.name
                for f in self.model._meta.fields
                if isinstance(f, AliasField) and f.name == original_fk_name
            ),
            original_fk_name,
        )

        return concrete_fk_name


class FieldsInline(FlexibleAdminMixin, StackedInline):
    """An inline representing a single field on a Form."""

    classes = ("collapse",)
    model = BaseField
    extra = 1

    fieldsets = (
        (None, {"fields": (("label", "field_type", "required"),)}),
        (
            "ADVANCED FIELD OPTIONS",
            {
                "classes": ("collapse",),
                "fields": (
                    "name",
                    "help_text",
                    "label_suffix",
                    "field_type_options",
                    "form_field_options",
                    "form_widget_options",
                ),
            },
        ),
    )

    @property
    def fk_name(self) -> str:
        """Return the name of the field with a ForeignKey to the Form.

        Handles the case where the implementer has aliased the "form" field.

        Returns:
            str: The name of the form ForeignKey field.
        """
        original_fk_name = BaseField.FlexibleMeta.form_field_name
        concrete_fk_name = next(  # pragma: no cover
            (
                f.original_field.name
                for f in self.model._meta.fields
                if isinstance(f, AliasField) and f.name == original_fk_name
            ),
            original_fk_name,
        )

        return concrete_fk_name

    @property
    def inlines(self) -> Iterable[InlineModelAdmin]:
        """Return valid inlines for the given BaseField implementation.

        Returns:
            Iterable[InlineModelAdmin]: An iterable of InlineModelAdmin classes.
        """
        return (
            *super().inlines,
            cast(
                FieldModifiersInline,
                type(
                    "FieldModifiersInline",
                    (FieldModifiersInline,),
                    {
                        "model": self.model._flexible_model_for(BaseFieldModifier),
                    },
                ),
            ),
        )


class FieldsetItemsInline(FlexibleAdminMixin, TabularInline):
    """An inline representing an item in a fieldset on a Form."""

    model = BaseFieldsetItem
    extra = 1

    @property
    def fk_name(self) -> str:
        """Return the name of the field with a ForeignKey to the Fieldset.

        Handles the case where the implementer has aliased the "fieldset" field.

        Returns:
            str: The name of the fieldset ForeignKey field.
        """
        original_fk_name = BaseFieldsetItem.FlexibleMeta.fieldset_field_name
        concrete_fk_name = next(  # pragma: no cover
            (
                f.original_field.name
                for f in self.model._meta.fields
                if isinstance(f, AliasField) and f.name == original_fk_name
            ),
            original_fk_name,
        )

        return concrete_fk_name

    def formfield_for_foreignkey(
        self, db_field: models.Field, request: HttpRequest, **kwargs: Any
    ) -> forms.Field:
        """Modify ForeignKey fields before they're rendered to the form.

        Restricts the "field" choices to only include Fields for the current
        Form.

        Args:
            db_field: The model field to be rendered to the form.
            request: The current HTTP request.
            kwargs: Passed to super.

        Returns:
            forms.Field: A configured form field for the given db_field.
        """
        if db_field.name == "field":
            form_id = request.resolver_match.kwargs.get("object_id")
            Field = self.model._flexible_model_for(BaseField)
            kwargs["queryset"] = (
                Field._default_manager.filter(form=form_id)
                if form_id
                else Field._default_manager.none()
            )
        return cast(
            forms.Field, super().formfield_for_foreignkey(db_field, request, **kwargs)
        )


class FieldsetsInline(FlexibleAdminMixin, StackedInline):
    """An inline representing a fieldset on a Form."""

    classes = ("collapse",)
    model = BaseFieldset
    extra = 1

    @property
    def fk_name(self) -> str:
        """Return the name of the field with a ForeignKey to the Form.

        Handles the case where the implementer has aliased the "form" field.

        Returns:
            str: The name of the form ForeignKey field.
        """
        original_fk_name = BaseFieldset.FlexibleMeta.form_field_name
        concrete_fk_name = next(  # pragma: no cover
            (
                f.original_field.name
                for f in self.model._meta.fields
                if isinstance(f, AliasField) and f.name == original_fk_name
            ),
            original_fk_name,
        )

        return concrete_fk_name

    @property
    def inlines(self) -> Iterable[InlineModelAdmin]:
        """Return valid inlines for the given BaseFieldset implementation.

        Returns:
            Iterable[InlineModelAdmin]: An iterable of InlineModelAdmin classes.
        """
        return (
            *super().inlines,
            cast(
                FieldsetItemsInline,
                type(
                    "FieldsetItemsInline",
                    (FieldsetItemsInline,),
                    {
                        "model": self.model._flexible_model_for(
                            FieldsetItemsInline.model
                        ),
                    },
                ),
            ),
        )


class FormsAdmin(FlexibleAdminMixin, ModelAdmin):
    """An admin configuration for managing flexible forms."""

    class Meta:
        pass

    list_display = ("label", "_fields_count", "_records_count", "_add_record")

    @property
    def inlines(self) -> Iterable[InlineModelAdmin]:
        """Return valid inlines for the given BaseForm implementation.

        Returns:
            Iterable[InlineModelAdmin]: An iterable of InlineModelAdmin classes.
        """
        return (
            *super().inlines,
            cast(
                FieldsInline,
                type(
                    "FieldsInline",
                    (FieldsInline,),
                    {"model": self.model._flexible_model_for(FieldsInline.model)},
                ),
            ),
            cast(
                FieldsetsInline,
                type(
                    "FieldsetsInline",
                    (FieldsetsInline,),
                    {"model": self.model._flexible_model_for(FieldsetsInline.model)},
                ),
            ),
        )

    def get_queryset(self, *args: Any, **kwargs: Any) -> "models.QuerySet[BaseForm]":
        """Overrides the default queryset to optimize fetches.

        Args:
            args: (Passed to super)
            kwargs: (Passed to super)

        Returns:
            models.QuerySet[Form]: An optimized queryset.
        """
        return cast(
            "models.QuerySet[BaseForm]",
            (
                super()
                .get_queryset(*args, **kwargs)
                .annotate(models.Count("records", distinct=True))
                .annotate(models.Count("fields", distinct=True))
            ),
        )

    def _fields_count(self, form: BaseForm) -> int:
        """The number of fields related to this form.

        Args:
            form: The form object.

        Returns:
            int: The number of fields on the form.
        """
        return form.fields__count  # type: ignore

    _fields_count.short_description = "Fields"  # type: ignore
    _fields_count.admin_order_field = "fields__count"  # type: ignore

    def _records_count(self, form: BaseForm) -> SafeText:
        """The number of records related to this form.

        Args:
            form: The form object.

        Returns:
            SafeText: The number of records related to the form in a
                hyperlink to the records listing with a filter for the form.
        """
        Record = form._flexible_model_for(BaseRecord)
        app_label = Record._meta.app_label  # noqa: WPS437
        model_name = Record._meta.model_name  # noqa: WPS437

        changelist_url = reverse(f"admin:{app_label}_{model_name}_changelist")
        filtered_changelist_url = f"{changelist_url}?form_id={form.pk}"
        record_count = form.records__count  # type: ignore

        return mark_safe(
            f'<a href="{filtered_changelist_url}">{record_count}</a>'
        )  # noqa: S308, S703, E501

    _records_count.short_description = "Records"  # type: ignore
    _records_count.admin_order_field = "records__count"  # type: ignore

    def _add_record(self, form: BaseForm) -> SafeText:
        Record = form._flexible_model_for(BaseRecord)
        app_label = Record._meta.app_label  # noqa: WPS437
        model_name = Record._meta.model_name  # noqa: WPS437
        form_field_name = Record.FlexibleMeta.form_field_name

        add_url = (
            reverse(
                f"admin:{app_label}_{model_name}_add",
            )
            + f"?{form_field_name}={form.pk}"
        )

        return mark_safe(
            f'<a href="{add_url}">Add record</a>'
        )  # noqa: S308, S703, E501


class RecordsAdmin(FlexibleAdminMixin, ModelAdmin):
    """An admin configuration for managing records."""

    class Meta:
        pass

    list_display = ("__str__", "_form_label")

    # Type hints.
    model: Type[BaseRecord]

    def get_queryset(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> "models.QuerySet[BaseRecord]":
        """Overrides the default queryset to optimize fetches.

        Args:
            args: (Passed to super)
            kwargs: (Passed to super)

        Returns:
            models.QuerySet[BaseRecord]: An optimized queryset.
        """
        return cast(
            "models.QuerySet[BaseRecord]",
            (
                super()
                .get_queryset(*args, **kwargs)
                .select_related("form")
                .prefetch_related(
                    "form__fields__modifiers",
                    "form__fieldsets__items__field",
                    "attributes__field",
                )
            ),
        )

    def _form_label(self, record: BaseRecord) -> SafeText:
        """Return the label of the record's form.

        Links to the change form for the form.

        Args:
            record: The record for which to render the form label.

        Returns:
            SafeText: The label of the record's form, linked to the change
                page for that form.
        """
        Form = record.form._flexible_model_for(BaseForm)
        app_label = Form._meta.app_label  # noqa: WPS437
        model_name = Form._meta.model_name  # noqa: WPS437

        change_url = reverse(
            f"admin:{app_label}_{model_name}_change", args=(record.form.pk,)
        )

        return mark_safe(
            f'<a href="{change_url}">{record.form.label}</a>'
        )  # noqa: S308, S703, E501

    _form_label.short_description = "Form"  # type: ignore
    _form_label.admin_order_field = "form__label"  # type: ignore

    def get_fieldsets(
        self,
        request: HttpRequest,
        obj: Optional[models.Model] = None,
    ) -> Sequence[DjangoFieldset]:
        """Return the fieldset configuration for the form.

        If the form has a fieldset configuration, use it instead of the
        default.
        """
        record_model = self.model._flexible_model_for(BaseRecord)
        form_model = self.model._flexible_model_for(BaseForm)
        form_field_name = record_model.FlexibleMeta.form_field_name
        form_pk = request.GET.get(form_field_name)

        default_fieldsets = cast(
            Sequence[DjangoFieldset],
            super().get_fieldsets(request, obj),
        )

        # Fieldsets starts out as the fieldsets Django builds by default.
        fieldsets = default_fieldsets

        if obj:
            obj = cast("BaseRecord", obj)
            fieldsets = obj.as_django_fieldsets() or default_fieldsets
        # If a form is specified in the query parameters, we'll look it up and
        # use its fieldset configuration until we have a record.
        elif form_pk:
            form_obj = cast(BaseForm, form_model.objects.get(pk=form_pk))
            fieldsets = form_obj.as_django_fieldsets() or default_fieldsets

        if fieldsets is not default_fieldsets:
            form = self.get_form(request, obj)
            form_fields = frozenset(form.base_fields.keys())  # type: ignore
            record_fields = tuple(
                f.name
                for f in self.model._meta.get_fields(include_parents=True)
                if f.name in form_fields
            )

            # If the record model has top-level attributes (in addition to
            # dynamic form attributes), add a metadata fieldset for managing
            # them.
            if record_fields:
                fieldsets = [
                    *fieldsets,
                    ("METADATA", {"fields": record_fields}),
                ]

        return fieldsets

    def get_form(
        self,
        request: HttpRequest,
        obj: Optional[models.Model] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Type[forms.BaseForm]:
        """Return the Django Form definition for the record.

        Generated dynamically if the Record has a form defined.

        Args:
            request: The current HTTP request.
            obj: The record for which to render the form.
            args: (Passed to super)
            kwargs: (Passed to super)

        Returns:
            forms.Form: The form object to be rendered.
        """
        if obj:
            return type(
                cast(BaseRecord, obj).as_django_form(
                    data=request.POST,
                    files=request.FILES,
                )
            )

        record_model = self.model._flexible_model_for(BaseRecord)
        form_model = self.model._flexible_model_for(BaseForm)
        form_field_name = record_model.FlexibleMeta.form_field_name
        form_pk = request.GET.get(form_field_name)

        if form_pk:
            form = cast(BaseForm, form_model.objects.get(pk=form_pk))
            django_form = type(form.as_django_form())

            # initial_values = {
            #     cast(str, f.name): request.GET[cast(str, f.name)]
            #     for f in (*record_model._meta.get_fields(), *form.fields.all())
            #     if f.name in request.GET
            #     and f.name not in (form_field_name,)
            # }
            # for field_name, initial_value in initial_values:
            #     django_form.fields[field_name].initial = initial_value

            return django_form

        # If no object was given and no form was specified, return a vanilla
        # model form for a BaseRecord.
        return cast(
            Type[forms.BaseForm],
            super().get_form(request, obj, *args, **kwargs),
        )
